import sys
import types

import numpy as np
import pytest
from sklearn.datasets import make_classification

from tabnetics.classification.backends import (
    ClassifierBackend,
    FLAMLBackend,
    OptunaBackend,
    SklearnBackend,
)


def _toy_data(n_samples=80, n_features=20, n_classes=3, seed=0):
    X, y = make_classification(
        n_samples=n_samples,
        n_features=n_features,
        n_classes=n_classes,
        n_informative=max(6, n_classes * 2),
        n_redundant=2,
        n_clusters_per_class=1,
        random_state=seed,
    )
    classes, counts = np.unique(y, return_counts=True)
    return np.asarray(X, dtype=float), np.asarray(y), int(classes.size), np.asarray(counts, dtype=int)


def test_classifier_backend_abc_cannot_be_instantiated():
    with pytest.raises(TypeError):
        _ = ClassifierBackend()  # type: ignore[abstract]


def test_supports_dataset_boundary_checks():
    backend = SklearnBackend(candidate_names=("lr",))
    assert backend.supports_dataset(n_samples=2, n_features=1, n_classes=2) is True
    assert backend.supports_dataset(n_samples=1, n_features=1, n_classes=2) is False
    assert backend.supports_dataset(n_samples=2, n_features=0, n_classes=2) is False
    assert backend.supports_dataset(n_samples=2, n_features=1, n_classes=1) is False


def test_sklearn_backend_basic_fit_and_select_smoke():
    X, y, n_classes, counts = _toy_data(n_samples=90, n_features=24, n_classes=3, seed=2)
    backend = SklearnBackend(candidate_names=("lr", "svm_rbf", "nb"))
    model, name, score, std, n_splits, meta = backend.fit_and_select(
        X, y, seed=7, n_classes=n_classes, class_counts=counts
    )
    assert name in {"lr", "svm_rbf", "nb"}
    assert np.isfinite(score) or np.isnan(score)
    assert np.isfinite(std) or np.isnan(std)
    assert int(n_splits) >= 0
    assert "classification_backend_used" in meta
    assert model is not None


def test_sklearn_backend_hybrid_scoring_path():
    X, y, n_classes, counts = _toy_data(n_samples=84, n_features=18, n_classes=3, seed=3)
    backend = SklearnBackend(
        candidate_names=("lr", "svm_rbf"),
        use_hybrid_score=True,
        hybrid_balanced_weight=0.7,
        hybrid_macro_f1_weight=0.3,
    )
    _, _, score, _, n_splits, _ = backend.fit_and_select(
        X, y, seed=11, n_classes=n_classes, class_counts=counts
    )
    assert (np.isfinite(score) and 0.0 <= score <= 1.0) or np.isnan(score)
    assert int(n_splits) >= 0


def test_sklearn_backend_expanded_classifier_pool_smoke():
    X, y, n_classes, counts = _toy_data(n_samples=120, n_features=30, n_classes=3, seed=4)
    backend = SklearnBackend(
        candidate_names=("nsc", "pls_da_classifier", "shrinkage_lda", "gpc", "lr"),
    )
    _, name, _, _, _, meta = backend.fit_and_select(
        X, y, seed=13, n_classes=n_classes, class_counts=counts
    )
    assert name in {"nsc", "pls_da_classifier", "shrinkage_lda", "gpc", "lr"}
    assert "model_cv_evaluated_candidates" in meta


def test_sklearn_backend_gpc_gated_for_large_n():
    X, y, n_classes, counts = _toy_data(n_samples=240, n_features=25, n_classes=3, seed=5)
    backend = SklearnBackend(candidate_names=("gpc", "lr"))
    _, name, _, _, _, _ = backend.fit_and_select(
        X, y, seed=17, n_classes=n_classes, class_counts=counts
    )
    # gpc is gated at n<=200; lr remains as safe candidate.
    assert name == "lr"


def test_flaml_supports_dataset_thresholds():
    backend = FLAMLBackend(
        min_n_for_automl=50,
        min_n_per_class_for_automl=10,
        max_p_over_n_for_automl=200,
    )
    counts_ok = np.array([25, 25], dtype=int)
    counts_low = np.array([4, 45], dtype=int)
    assert backend.supports_dataset(n_samples=49, n_features=20, n_classes=2, class_counts=counts_ok) is False
    assert backend.supports_dataset(n_samples=50, n_features=20, n_classes=2, class_counts=counts_ok) is True
    assert backend.supports_dataset(n_samples=60, n_features=20000, n_classes=2, class_counts=np.array([30, 30])) is False
    assert backend.supports_dataset(n_samples=49, n_features=20, n_classes=2, class_counts=counts_low) is False


def test_flaml_fit_and_select_returns_lr_when_cv_too_small():
    X, y, n_classes, _ = _toy_data(n_samples=60, n_features=10, n_classes=2, seed=6)
    class_counts = np.array([3, 57], dtype=int)  # min class count < min_n_per_class_for_cv
    backend = FLAMLBackend(min_n_per_class_for_cv=5)
    model, name, score, std, n_splits, meta = backend.fit_and_select(
        X,
        y,
        seed=19,
        n_classes=n_classes,
        class_counts=class_counts,
    )
    assert name == "lr"
    assert int(n_splits) == 0
    assert np.isnan(score)
    assert np.isnan(std)
    assert meta.get("classification_guard_reason") == "min_n_per_class_for_cv"
    assert model is not None


def test_flaml_fit_and_select_raises_import_error_when_missing():
    X, y, n_classes, counts = _toy_data(n_samples=90, n_features=10, n_classes=2, seed=7)
    backend = FLAMLBackend()
    with pytest.raises(ImportError):
        backend.fit_and_select(X, y, seed=23, n_classes=n_classes, class_counts=counts)


def test_flaml_metric_mapping():
    assert FLAMLBackend._map_scoring("balanced_accuracy", "macro_f1") == "accuracy"
    assert FLAMLBackend._map_scoring("balanced-accuracy", "macro_f1") == "accuracy"
    assert FLAMLBackend._map_scoring("bal_acc", "macro_f1") == "accuracy"
    assert FLAMLBackend._map_scoring("macro_f1", "accuracy") == "macro_f1"
    assert FLAMLBackend._map_scoring("unknown_metric", "balanced_accuracy") == "accuracy"
    assert FLAMLBackend._map_scoring("unknown_metric", "macro_f1") == "macro_f1"
    assert FLAMLBackend(metric="balanced_accuracy").metric == "accuracy"


def test_flaml_happy_path_with_fake_module(monkeypatch):
    X, y, n_classes, counts = _toy_data(n_samples=90, n_features=12, n_classes=2, seed=9)

    class _FakeAutoML:
        def __init__(self):
            self.best_loss = 0.22
            self.best_estimator = "rf"
            self.best_config = {"max_depth": 5}
            self.model = object()

        def fit(self, *args, **kwargs):
            return None

    fake_module = types.SimpleNamespace(AutoML=_FakeAutoML)
    monkeypatch.setitem(sys.modules, "flaml", fake_module)

    backend = FLAMLBackend(time_budget=17)
    model, name, score, std, n_splits, meta = backend.fit_and_select(
        X, y, seed=29, n_classes=n_classes, class_counts=counts
    )
    assert name == "flaml_rf"
    assert model is not None
    assert score == pytest.approx(0.78)
    assert np.isnan(std)
    assert int(n_splits) >= 2
    assert meta.get("flaml_best_estimator") == "rf"


def test_optuna_supports_dataset_thresholds():
    backend = OptunaBackend(
        candidate_names=("lr", "svm_rbf"),
        min_n_for_automl=50,
        min_n_per_class_for_automl=10,
        max_p_over_n_for_automl=200,
    )
    counts_ok = np.array([25, 25], dtype=int)
    counts_low = np.array([4, 45], dtype=int)
    assert backend.supports_dataset(n_samples=49, n_features=20, n_classes=2, class_counts=counts_ok) is False
    assert backend.supports_dataset(n_samples=50, n_features=20, n_classes=2, class_counts=counts_ok) is True
    assert backend.supports_dataset(n_samples=60, n_features=20000, n_classes=2, class_counts=np.array([30, 30])) is False
    assert backend.supports_dataset(n_samples=49, n_features=20, n_classes=2, class_counts=counts_low) is False


def test_optuna_fit_and_select_returns_lr_when_cv_too_small():
    X, y, n_classes, _ = _toy_data(n_samples=60, n_features=10, n_classes=2, seed=12)
    class_counts = np.array([3, 57], dtype=int)  # min class count < min_n_per_class_for_cv
    backend = OptunaBackend(candidate_names=("lr", "svm_rbf"), min_n_per_class_for_cv=5)
    model, name, score, std, n_splits, meta = backend.fit_and_select(
        X,
        y,
        seed=19,
        n_classes=n_classes,
        class_counts=class_counts,
    )
    assert name == "lr"
    assert int(n_splits) == 0
    assert np.isnan(score)
    assert np.isnan(std)
    assert meta.get("classification_guard_reason") == "min_n_per_class_for_cv"
    assert model is not None


def test_optuna_fit_and_select_raises_import_error_when_missing(monkeypatch):
    X, y, n_classes, counts = _toy_data(n_samples=90, n_features=10, n_classes=2, seed=13)
    monkeypatch.setitem(sys.modules, "optuna", None)
    backend = OptunaBackend(candidate_names=("lr", "svm_rbf"))
    with pytest.raises(ImportError):
        backend.fit_and_select(X, y, seed=23, n_classes=n_classes, class_counts=counts)


def test_optuna_happy_path_with_fake_module(monkeypatch):
    X, y, n_classes, counts = _toy_data(n_samples=90, n_features=12, n_classes=2, seed=14)

    class _FakeTrial:
        def __init__(self):
            self.params = {}

        def suggest_categorical(self, name, choices):
            value = choices[0]
            self.params[name] = value
            return value

        def suggest_float(self, name, low, high, log=False):
            _ = (high, log)
            value = float(low)
            self.params[name] = value
            return value

        def suggest_int(self, name, low, high):
            _ = high
            value = int(low)
            self.params[name] = value
            return value

    class _FakeStudy:
        def __init__(self):
            self.trials = []
            self.best_trial = None

        def optimize(self, objective, n_trials=10, timeout=None, n_jobs=1):
            _ = (timeout, n_jobs)
            best_val = -np.inf
            for _ in range(int(max(1, n_trials))):
                trial = _FakeTrial()
                val = float(objective(trial))
                trial.value = val
                self.trials.append(trial)
                if self.best_trial is None or val > best_val:
                    self.best_trial = trial
                    best_val = val

    class _FakeTPESampler:
        def __init__(self, seed=0):
            self.seed = int(seed)

    fake_module = types.SimpleNamespace(
        create_study=lambda direction, sampler: _FakeStudy(),
        samplers=types.SimpleNamespace(TPESampler=_FakeTPESampler),
    )
    monkeypatch.setitem(sys.modules, "optuna", fake_module)

    backend = OptunaBackend(
        candidate_names=("lr", "svm_linear"),
        time_budget=15,
        n_trials=2,
        min_n_for_automl=10,
        min_n_per_class_for_automl=2,
        min_n_per_class_for_cv=2,
        max_p_over_n_for_automl=10_000,
    )
    model, name, score, std, n_splits, meta = backend.fit_and_select(
        X,
        y,
        seed=31,
        n_classes=n_classes,
        class_counts=counts,
    )
    assert name.startswith("optuna_")
    assert model is not None
    assert np.isfinite(score) or np.isnan(score)
    assert np.isfinite(std) or np.isnan(std)
    assert int(n_splits) >= 2
    assert meta.get("classification_backend_used") == "optuna"



import io
import os
import numpy as np
import pandas as pd
import pytest
from pathlib import Path
from types import SimpleNamespace
from scipy.io import savemat

from tabnetics.validation.suite import (
    CATALOG,
    COMPONENT_DEFAULTS,
    _apply_scenario_defaults_to_configs,
    _dataset_promotion_metadata,
    _load_dense_arff_dataset,
    _load_manual_tabular_dataset,
    _generate_distribution_cases,
    _integrated_scenario_defaults,
    _project_components_for_pipeline,
    _safe_train_test_split,
    _summarize_runs,
    AblationConfig,
    DatasetIntegrityPolicyError,
    DatasetIntegritySkipError,
    ValidationDatasetSpec,
    build_arg_parser,
    build_ablation_configs,
    load_feature_selection_dataset,
    run_validation_suite,
    resolve_dataset_ids,
)


def _has_real_source_for(spec):
    """Return True when manual/HF real data source is configured for this spec."""
    if os.environ.get("TABNETICS_HF_ORG", "").strip():
        return True

    params = dict(spec.params or {})
    env_name = str(params.get("local_path_env", "")).strip()
    if env_name:
        env_path = os.environ.get(env_name, "").strip()
        if env_path and Path(env_path).exists():
            return True

    repo_root = Path(__file__).resolve().parents[1]
    candidates = []
    default_one = params.get("default_local_path")
    if default_one:
        candidates.append(str(default_one))
    for extra in list(params.get("default_local_paths", []) or ()):
        candidates.append(str(extra))

    for cand in candidates:
        p = Path(cand)
        if not p.is_absolute():
            p = repo_root / p
        if p.exists():
            return True
    return False


def test_resolve_dataset_ids_pipeline_and_set_filtering():
    selected = resolve_dataset_ids(
        catalog=CATALOG,
        dataset_sets=["fs_easy"],
        explicit_ids=[],
        exclude_ids=["dlbcl_shipp"],
        pipelines=["fs"],
        max_datasets=None,
    )

    assert "leukemia_golub" in selected
    assert "dlbcl_shipp" not in selected
    assert all(CATALOG[ds_id].pipeline == "fs" for ds_id in selected)


def test_build_ablation_configs_single_off_subset():
    configs = build_ablation_configs(
        profile="single_off",
        base_components=dict(COMPONENT_DEFAULTS),
        pipelines=["fs"],
        constrained_components=["fs.tritrust", "fs.method_mrmr_jmi"],
    )

    names = [cfg.name for cfg in configs]
    assert names[0] == "baseline"
    assert "disable_fs_tritrust" in names
    assert "disable_fs_method_mrmr_jmi" in names
    assert len(configs) == 3


def test_manual_or_synth_dataset_falls_back_to_synthetic_when_env_missing(monkeypatch):
    # Use a manual dataset and hide both env var *and* default_local_path so the
    # loader reliably falls back to synthetic.
    import copy

    spec = copy.deepcopy(CATALOG["cumida_leukemia_subtypes"])
    monkeypatch.delenv("CUMIDA_LEUKEMIA_PATH", raising=False)
    # Remove default_local_path so the repo-local fallback path isn't tried.
    spec.params.pop("default_local_path", None)
    spec.params.pop("default_local_paths", None)
    loaded = load_feature_selection_dataset(
        spec,
        seed=11,
        allow_synthetic_fallback=True,
        sample_cap=120,
        feature_cap=300,
    )

    assert loaded.data_source == "synthetic_fallback"
    assert loaded.X.shape[0] <= 120
    assert loaded.X.shape[1] <= 300
    assert loaded.X.shape[0] == loaded.y.shape[0]
    assert np.issubdtype(loaded.X.dtype, np.number)


def test_validation_suite_cli_defaults_are_strict():
    args = build_arg_parser().parse_args([])
    assert args.allow_synthetic_fallback is False
    assert str(args.dataset_integrity_policy) == "error"


def test_require_hf_source_rejects_missing_hf_bundle(monkeypatch):
    import copy

    spec = copy.deepcopy(CATALOG["cumida_brain_gse50161"])
    monkeypatch.delenv("TABNETICS_HF_ORG", raising=False)
    monkeypatch.delenv("TABNETICS_HF_REPO_ID", raising=False)

    with pytest.raises(RuntimeError, match="HuggingFace bundle"):
        load_feature_selection_dataset(
            spec,
            seed=11,
            allow_synthetic_fallback=False,
            sample_cap=120,
            feature_cap=300,
            source_policy="real_only",
            require_hf_source=True,
        )


def test_validation_suite_rejects_allow_synthetic_fallback(monkeypatch, tmp_path):
    monkeypatch.setenv("TABNETICS_HF_ORG", "test-org")
    monkeypatch.delenv("TABNETICS_HF_REPO_ID", raising=False)

    args = build_arg_parser().parse_args(
        [
            "--datasets",
            "cumida_brain_gse50161",
            "--allow-synthetic-fallback",
            "--output-dir",
            str(tmp_path),
        ]
    )

    with pytest.raises(ValueError, match="Synthetic fallback is forbidden"):
        run_validation_suite(args)


def test_face_proxy_loader_builds_face_domain_dataset(monkeypatch):
    rng = np.random.default_rng(7)
    fake_images = rng.normal(loc=0.5, scale=0.2, size=(400, 64, 64)).astype(float)
    fake_images = np.clip(fake_images, 0.0, 1.0)
    fake_target = np.repeat(np.arange(40, dtype=int), 10)

    def _fake_fetch_olivetti_faces(*, shuffle: bool = False, download_if_missing: bool = True):
        return SimpleNamespace(images=fake_images, target=fake_target)

    monkeypatch.setattr("tabnetics.domains.face.datasets.fetch_olivetti_faces", _fake_fetch_olivetti_faces)

    spec = CATALOG["warp_pie10p"]
    loaded = load_feature_selection_dataset(
        spec,
        seed=11,
        allow_synthetic_fallback=False,
        sample_cap=500,
        feature_cap=5000,
    )

    assert loaded.data_source.startswith("sklearn:olivetti_faces_proxy:")
    assert loaded.X.shape == (210, 2420)
    assert loaded.y.shape == (210,)
    assert len(np.unique(loaded.y)) == 10


def test_mat_url_loader_builds_dataset_from_downloaded_payload(monkeypatch):
    X_ref = np.asarray(
        [
            [1.0, 2.0, 3.0],
            [4.0, 5.0, 6.0],
            [7.0, 8.0, 9.0],
            [10.0, 11.0, 12.0],
        ],
        dtype=float,
    )
    y_ref = np.asarray([1, 1, 2, 3], dtype=int).reshape(-1, 1)
    buf = io.BytesIO()
    savemat(buf, {"X": X_ref, "Y": y_ref})
    payload = buf.getvalue()

    monkeypatch.setattr("tabnetics.validation.suite._download_url_bytes", lambda _url, timeout_sec=60: payload)

    spec = ValidationDatasetSpec(
        dataset_id="toy_mat_url_loader",
        display_name="Toy MAT URL",
        pipeline="fs",
        tier="hard",
        loader_kind="mat_url_or_synth",
        params={
            "mat_url_options": [{"url": "https://example.invalid/toy.mat"}],
            "synthetic_profile": {
                "n_samples": 40,
                "n_features": 8,
                "n_classes": 3,
                "difficulty": "hard",
            },
        },
    )

    loaded = load_feature_selection_dataset(
        spec,
        seed=11,
        allow_synthetic_fallback=False,
        sample_cap=100,
        feature_cap=100,
    )

    assert loaded.data_source.startswith("mat_url:https://example.invalid/toy.mat")
    assert loaded.X.shape == (4, 3)
    assert loaded.y.shape == (4,)
    assert len(np.unique(loaded.y)) == 3


def test_nci60_proxy_loader_materializes_61x4_with_expected_class_count(monkeypatch):
    labels = (
        ["RENAL"] * 9
        + ["NSCLC"] * 9
        + ["MELANOMA"] * 8
        + ["BREAST"] * 7
        + ["COLON"] * 7
        + ["OVARIAN"] * 6
        + ["LEUKEMIA"] * 6
        + ["CNS"] * 5
        + ["PROSTATE"] * 2
        + ["UNKNOWN"] * 1
        + ["K562A-repro"] * 1
        + ["K562B-repro"] * 1
        + ["MCF7A-repro"] * 1
        + ["MCF7D-repro"] * 1
    )
    assert len(labels) == 64

    rng = np.random.default_rng(7)
    features = rng.normal(size=(64, 10))
    df = pd.DataFrame(features, columns=[f"g{i}" for i in range(10)])
    df.insert(0, "Unnamed: 0", np.arange(64))
    df["label"] = labels

    monkeypatch.setattr("tabnetics.validation.suite.pd.read_csv", lambda _url: df)

    spec = ValidationDatasetSpec(
        dataset_id="toy_nci60_proxy",
        display_name="Toy NCI60 Proxy",
        pipeline="fs",
        tier="hard",
        loader_kind="nci60_proxy_or_synth",
        params={
            "proxy_target_features": 8,
            "synthetic_profile": {
                "n_samples": 61,
                "n_features": 8,
                "n_classes": 8,
                "difficulty": "hard",
            },
        },
    )

    loaded = load_feature_selection_dataset(
        spec,
        seed=11,
        allow_synthetic_fallback=False,
        sample_cap=200,
        feature_cap=200,
    )

    assert loaded.data_source.startswith("rdatasets:ISLR:NCI60:proxy_")
    assert loaded.X.shape == (61, 8)
    assert loaded.y.shape == (61,)
    assert len(np.unique(loaded.y)) == 8
    assert "nci60_proxy_transform" in (loaded.notes or "")


def test_cumida_brain_loads_real_data_when_available():
    """CuMiDa Brain should load real data when available; synthetic fallback is optional."""
    spec = CATALOG["cumida_brain_gse50161"]
    real_available = _has_real_source_for(spec)
    loaded = load_feature_selection_dataset(
        spec,
        seed=19,
        allow_synthetic_fallback=True,
        sample_cap=120,
        feature_cap=300,
    )

    if real_available:
        assert loaded.data_source.startswith(("manual_arff:", "hf_bundle:"))
    else:
        assert loaded.data_source == "synthetic_fallback"
    assert loaded.X.shape[0] == loaded.y.shape[0]
    # Synthetic profile mirrors 4-class real-data topology.
    assert len(np.unique(loaded.y)) == 4


def test_cumida_brain_loads_real_data_when_fallback_disabled():
    spec = CATALOG["cumida_brain_gse50161"]
    if _has_real_source_for(spec):
        loaded = load_feature_selection_dataset(
            spec,
            seed=19,
            allow_synthetic_fallback=False,
            sample_cap=120,
            feature_cap=300,
        )
        assert loaded.data_source.startswith(("manual_arff:", "hf_bundle:"))
        return

    try:
        load_feature_selection_dataset(
            spec,
            seed=19,
            allow_synthetic_fallback=False,
            sample_cap=120,
            feature_cap=300,
        )
    except RuntimeError as exc:
        msg = str(exc)
        assert "No data source for dataset=cumida_brain_gse50161" in msg
    else:
        raise AssertionError("Expected missing-source RuntimeError when no real CuMiDa source is configured.")


def test_cumida_brain_promotion_metadata_is_promotable():
    """CuMiDa brain is a real-data benchmark; promotion metadata should not mark it fallback-only."""
    spec = CATALOG["cumida_brain_gse50161"]
    meta = _dataset_promotion_metadata(spec)

    assert int(meta["promotion_eligible"]) == 1
    assert str(meta["promotion_blocker"]) == ""
    assert str(meta["source_policy"]) in {"standard", "real_only"}


def test_dense_arff_loader_parses_numeric_features_and_label(tmp_path):
    content = """@RELATION toy

@ATTRIBUTE f1 NUMERIC
@ATTRIBUTE f2 NUMERIC
@ATTRIBUTE class {a,b}

@DATA
1.0,2.0,a
3.5,4.25,b
"""
    path = tmp_path / "toy.arff"
    path.write_text(content, encoding="utf-8")

    loaded = _load_dense_arff_dataset(path)
    assert loaded.X.shape == (2, 2)
    assert loaded.y.shape == (2,)
    assert loaded.data_source.startswith("manual_arff:")


def test_summarize_runs_preserves_domain_and_platform_dimensions():
    runs_df = pd.DataFrame(
        [
            {
                "pipeline": "fs",
                "dataset_id": "leukemia_golub",
                "config_name": "baseline",
                "domain": "genomics",
                "platform": "Affy HG-U95",
                "accuracy": 0.90,
                "balanced_accuracy": 0.91,
                "macro_f1": 0.89,
                "hybrid_score": 0.90,
            },
            {
                "pipeline": "fs",
                "dataset_id": "orlraws10p",
                "config_name": "baseline",
                "domain": "non_genomic",
                "platform": "cDNA",
                "accuracy": 0.88,
                "balanced_accuracy": 0.87,
                "macro_f1": 0.86,
                "hybrid_score": 0.87,
            },
        ]
    )
    summary_df, by_config_df = _summarize_runs(runs_df)

    assert "domain" in summary_df.columns
    assert "platform" in summary_df.columns
    assert "pipeline" in by_config_df.columns


def test_dense_arff_loader_parses_question_mark_missing_values(tmp_path):
    content = """@RELATION toy

@ATTRIBUTE f1 NUMERIC
@ATTRIBUTE f2 NUMERIC
@ATTRIBUTE class {a,b}

@DATA
1.0,?,a
?,2.0,b
"""
    path = tmp_path / "toy_missing.arff"
    path.write_text(content, encoding="utf-8")

    loaded = _load_dense_arff_dataset(path)
    assert loaded.X.shape == (2, 2)
    assert np.isnan(loaded.X[0, 1])
    assert np.isnan(loaded.X[1, 0])


def test_manual_arff_loader_rejects_git_lfs_pointer_stub(tmp_path):
    content = """version https://git-lfs.github.com/spec/v1
oid sha256:0000000000000000000000000000000000000000000000000000000000000000
size 123
"""
    path = tmp_path / "pointer.arff"
    path.write_text(content, encoding="utf-8")

    try:
        _load_manual_tabular_dataset(path)
        raise AssertionError("Expected RuntimeError for Git LFS pointer file.")
    except RuntimeError as exc:
        assert "Git LFS pointer file" in str(exc)


def test_gisette_openml_option_prefers_active_version():
    spec = CATALOG["gisette_nips03"]
    options = list(spec.params.get("openml_options", []) or [])
    assert len(options) >= 1
    assert str(options[0].get("name", "")).lower() == "gisette"
    assert int(options[0].get("version", 0)) == 2


def test_dense_arff_loader_drops_unknown_declared_class_rows(tmp_path):
    content = """@RELATION toy

@ATTRIBUTE f1 NUMERIC
@ATTRIBUTE class {a,b}

@DATA
1.0,a
2.0,b
3.0,c
"""
    path = tmp_path / "toy_unknown_class.arff"
    path.write_text(content, encoding="utf-8")

    loaded = _load_dense_arff_dataset(path)
    assert loaded.X.shape == (2, 1)
    assert loaded.y.shape == (2,)
    assert "unknown_class_rows_dropped:1" in (loaded.notes or "")


def test_manual_dataset_integrity_policy_fallback_uses_synthetic(tmp_path):
    content = """@RELATION toy

@ATTRIBUTE f1 NUMERIC
@ATTRIBUTE f2 NUMERIC
@ATTRIBUTE class {a,b}

@DATA
1.0,2.0,a
2.0,3.0,a
3.0,4.0,a
"""
    path = tmp_path / "single_class.arff"
    path.write_text(content, encoding="utf-8")

    spec = ValidationDatasetSpec(
        dataset_id="toy_manual_single_class",
        display_name="Toy Manual Single Class",
        pipeline="fs",
        tier="hard",
        loader_kind="manual_or_synth",
        params={
            "default_local_path": str(path),
            "synthetic_profile": {
                "n_samples": 80,
                "n_features": 120,
                "n_classes": 3,
                "difficulty": "hard",
            },
        },
    )

    loaded = load_feature_selection_dataset(
        spec,
        seed=11,
        allow_synthetic_fallback=True,
        sample_cap=90,
        feature_cap=130,
        class_integrity_policy="fallback",
        class_min_classes=2,
        class_min_class_count=1,
    )

    classes = np.unique(loaded.y)
    assert loaded.data_source == "synthetic_fallback"
    assert loaded.X.shape[0] <= 90
    assert loaded.X.shape[1] <= 130
    assert classes.size >= 2
    assert "integrity_fallback_from:manual_arff:" in (loaded.notes or "")


def test_manual_dataset_integrity_policy_skip_raises(tmp_path):
    content = """@RELATION toy

@ATTRIBUTE f1 NUMERIC
@ATTRIBUTE class {a,b}

@DATA
1.0,a
2.0,a
3.0,a
"""
    path = tmp_path / "single_class_skip.arff"
    path.write_text(content, encoding="utf-8")

    spec = ValidationDatasetSpec(
        dataset_id="toy_manual_single_class_skip",
        display_name="Toy Manual Single Class Skip",
        pipeline="fs",
        tier="hard",
        loader_kind="manual_or_synth",
        params={
            "default_local_path": str(path),
            "synthetic_profile": {
                "n_samples": 80,
                "n_features": 120,
                "n_classes": 2,
                "difficulty": "hard",
            },
        },
    )

    try:
        load_feature_selection_dataset(
            spec,
            seed=13,
            allow_synthetic_fallback=True,
            sample_cap=100,
            feature_cap=140,
            class_integrity_policy="skip",
            class_min_classes=2,
            class_min_class_count=1,
        )
        raise AssertionError("Expected DatasetIntegritySkipError for single-class manual dataset.")
    except DatasetIntegritySkipError:
        pass


def test_distribution_case_generation_for_all_df_specs():
    df_specs = [spec for spec in CATALOG.values() if spec.pipeline == "df"]
    for spec in df_specs:
        cases = _generate_distribution_cases(spec, seed=7)
        assert len(cases) > 0, spec.dataset_id


def test_integrated_scenario_defaults_enable_expected_components():
    base = dict(COMPONENT_DEFAULTS)
    base_proj = _project_components_for_pipeline(base, "integrated")
    cfgs = [AblationConfig(name="baseline", components=base_proj)]

    spec = CATALOG["int_low_gof_downweighting"]
    scenario_defaults = _integrated_scenario_defaults(spec)
    adjusted = _apply_scenario_defaults_to_configs(cfgs, base_proj, scenario_defaults)

    assert adjusted[0].components["integrated.low_gof_downweighting"] is True


def test_safe_train_test_split_respects_max_train_samples_cap():
    rng = np.random.default_rng(17)
    X = rng.normal(size=(200, 8))
    y = np.asarray(([0, 1] * 100), dtype=int)

    X_train, X_test, y_train, y_test = _safe_train_test_split(
        X,
        y,
        test_size=0.20,
        seed=17,
        max_train_samples=60,
    )

    assert X_train.shape[0] == 60
    assert X_test.shape[0] == 140
    assert y_train.shape[0] == 60
    assert y_test.shape[0] == 140


def test_safe_train_test_split_enforces_minimum_80_20_holdout_when_capped():
    rng = np.random.default_rng(23)
    X = rng.normal(size=(40, 6))
    y = np.asarray(([0, 1] * 20), dtype=int)

    X_train, X_test, y_train, y_test = _safe_train_test_split(
        X,
        y,
        test_size=0.20,
        seed=23,
        max_train_samples=36,
    )

    assert X_train.shape[0] == 32
    assert X_test.shape[0] == 8
    assert y_train.shape[0] == 32
    assert y_test.shape[0] == 8



"""Tests for T-002: Multi-Model MNPO Oracles.

Validates that per-model performance oracle matrices are created correctly
when ``performance_oracle_mode="multi_model_oracles"`` (opt-in), and that
the default ``"single"`` mode is bit-for-bit backward compatible.
"""

import unittest
import numpy as np

from tabnetics.feature_selection.mnpo.oracles import (
    compute_oracle_stability_diagnostics,
    estimate_oracle_preferences,
    fit_tritrust_weights,
    pairwise_pref_from_fold_scores,
)
from tabnetics.core.mnpo import pairwise_pref_from_fold_scores as pairwise_pref_from_fold_scores_core
from tabnetics.feature_selection.mnpo.portfolio import evaluate_candidate_library
from tabnetics.feature_selection.config import (
    MNPOConfig,
    FeatureSelectorConfig,
    EvaluationConfig,
)


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _make_evaluation(n_candidates=4, n_folds=5, n_features=20, rng_seed=42,
                     include_per_model=False, model_keys=("lr_l2", "linear_svc", "rf_small")):
    """Build a synthetic evaluation dict for testing oracle construction."""
    rng = np.random.default_rng(rng_seed)
    names = [f"method_{i}" for i in range(n_candidates)]
    evaluation = {}
    for idx, name in enumerate(names):
        base_perf = 0.5 + 0.1 * idx  # monotonically increasing
        perf_scores = rng.normal(base_perf, 0.05, size=n_folds).clip(0, 1)
        selected = np.sort(rng.choice(n_features, size=min(5 + idx, n_features), replace=False))
        entry = {
            'performance_scores': np.asarray(perf_scores, dtype=float),
            'performance_mean': float(np.mean(perf_scores)),
            'stability': float(rng.uniform(0.3, 0.9)),
            'complexity': float(1.0 - len(selected) / n_features),
            'robustness': float(np.mean(perf_scores) - 0.02),
            'prediction_signal': rng.normal(0, 1, size=n_folds * 5),
            'target_signal': rng.integers(0, 2, size=n_folds * 5),
            'selected_indices': selected,
            'score_vector': rng.uniform(0, 1, size=n_features),
            'performance_scores_by_model': {},
        }
        if include_per_model:
            for mk in model_keys:
                mk_base = base_perf + rng.uniform(-0.05, 0.05)
                entry['performance_scores_by_model'][mk] = np.asarray(
                    rng.normal(mk_base, 0.04, size=n_folds).clip(0, 1), dtype=float
                )
        evaluation[name] = entry
    return names, evaluation


def test_pairwise_pref_from_fold_scores_canonical_matches_wrappers():
    from tabnetics.classification.backends import (
        _pairwise_pref_from_fold_scores as pairwise_pref_from_fold_scores_classification,
    )

    scores_i = np.asarray([0.70, 0.75, 0.73, np.nan, 0.77], dtype=float)
    scores_j = np.asarray([0.69, 0.76, 0.71, 0.72, np.nan], dtype=float)
    delta = 0.01

    expected = float(
        pairwise_pref_from_fold_scores_core(
            scores_i,
            scores_j,
            pairwise_delta=float(delta),
        )
    )
    got_oracles = float(pairwise_pref_from_fold_scores(scores_i, scores_j, pairwise_delta=float(delta)))
    got_classification = float(
        pairwise_pref_from_fold_scores_classification(
            scores_i,
            scores_j,
            pairwise_delta=float(delta),
        )
    )

    assert np.isclose(got_oracles, expected)
    assert np.isclose(got_classification, expected)


# ---------------------------------------------------------------------------
# Test Suite
# ---------------------------------------------------------------------------

class TestToggleOff(unittest.TestCase):
    """performance_oracle_mode='single': only the aggregate 'performance' oracle exists."""

    def test_single_mode_only_performance_oracle(self):
        names, evaluation = _make_evaluation(n_candidates=4, include_per_model=False)
        oracle_matrices, oracle_scores, _, _ = estimate_oracle_preferences(
            names, evaluation,
            pairwise_delta=0.01,
            use_tail_risk_oracle=False,
            tail_risk_alpha=0.33,
            use_qre_smoothing=False,
            qre_temperature_gamma=1.0,
            use_regret_oracle=False,
            use_stability_oracle=False,
            use_complexity_oracle=False,
            use_robust_oracle=False,
            use_diversity_oracle=False,
            diversity_oracle_mode="legacy_jaccard",
            diversity_redundancy_weight=0.6,
            diversity_complementarity_weight=0.35,
            performance_oracle_mode="single",
        )
        self.assertIn("performance", oracle_matrices)
        # No per-model oracles
        per_model_keys = [k for k in oracle_matrices if k.startswith("performance_")]
        self.assertEqual(len(per_model_keys), 0)

    def test_single_mode_ignores_per_model_data(self):
        """Even if per-model data is present, single mode does not create per-model oracles."""
        names, evaluation = _make_evaluation(n_candidates=3, include_per_model=True)
        oracle_matrices, _, _, _ = estimate_oracle_preferences(
            names, evaluation,
            pairwise_delta=0.01,
            use_tail_risk_oracle=False, tail_risk_alpha=0.33,
            use_qre_smoothing=False, qre_temperature_gamma=1.0,
            use_regret_oracle=False,
            use_stability_oracle=False, use_complexity_oracle=False,
            use_robust_oracle=False, use_diversity_oracle=False,
            diversity_oracle_mode="legacy_jaccard",
            diversity_redundancy_weight=0.6,
            diversity_complementarity_weight=0.35,
            performance_oracle_mode="single",
        )
        per_model_keys = [k for k in oracle_matrices if k.startswith("performance_")]
        self.assertEqual(len(per_model_keys), 0)


class TestToggleOn(unittest.TestCase):
    """performance_oracle_mode='multi_model_oracles': per-model oracle matrices exist."""

    def _get_oracles(self, n_candidates=4, model_keys=("lr_l2", "linear_svc", "rf_small")):
        names, evaluation = _make_evaluation(
            n_candidates=n_candidates, include_per_model=True, model_keys=model_keys,
        )
        return estimate_oracle_preferences(
            names, evaluation,
            pairwise_delta=0.01,
            use_tail_risk_oracle=False, tail_risk_alpha=0.33,
            use_qre_smoothing=False, qre_temperature_gamma=1.0,
            use_regret_oracle=False,
            use_stability_oracle=True, use_complexity_oracle=True,
            use_robust_oracle=False, use_diversity_oracle=False,
            diversity_oracle_mode="legacy_jaccard",
            diversity_redundancy_weight=0.6,
            diversity_complementarity_weight=0.35,
            performance_oracle_mode="multi_model_oracles",
        ), names

    def test_per_model_oracles_exist(self):
        (oracle_matrices, oracle_scores, _, _), names = self._get_oracles()
        self.assertIn("performance", oracle_matrices)
        self.assertIn("performance_lr_l2", oracle_matrices)
        self.assertIn("performance_linear_svc", oracle_matrices)
        self.assertIn("performance_rf_small", oracle_matrices)

    def test_oracle_matrix_shapes(self):
        """All per-model matrices are m×m."""
        (oracle_matrices, _, _, _), names = self._get_oracles(n_candidates=5)
        m = len(names)
        for key, mat in oracle_matrices.items():
            if key.startswith("performance"):
                self.assertEqual(mat.shape, (m, m), msg=f"Matrix {key} has wrong shape")

    def test_oracle_matrix_diagonal(self):
        """Diagonal of all oracle matrices = 0.5."""
        (oracle_matrices, _, _, _), names = self._get_oracles()
        for key, mat in oracle_matrices.items():
            if key.startswith("performance"):
                np.testing.assert_allclose(
                    np.diag(mat), 0.5,
                    err_msg=f"Diagonal of {key} is not 0.5",
                )

    def test_oracle_matrix_antisymmetric(self):
        """mat[i,j] + mat[j,i] = 1.0 for all off-diagonal entries."""
        (oracle_matrices, _, _, _), names = self._get_oracles()
        m = len(names)
        for key, mat in oracle_matrices.items():
            if key.startswith("performance"):
                for i in range(m):
                    for j in range(i + 1, m):
                        self.assertAlmostEqual(
                            mat[i, j] + mat[j, i], 1.0,
                            places=10,
                            msg=f"{key}[{i},{j}]+[{j},{i}] != 1.0",
                        )

    def test_oracle_scores_present(self):
        (_, oracle_scores, _, _), names = self._get_oracles()
        self.assertIn("performance_lr_l2", oracle_scores)
        self.assertEqual(oracle_scores["performance_lr_l2"].shape, (len(names),))


class TestTriTrustReference(unittest.TestCase):
    """TriTrust reference oracle selection with multi-model mode."""

    def test_reference_uses_lr_l2_when_available(self):
        """When performance_lr_l2 is present, TriTrust should use it as reference."""
        names, evaluation = _make_evaluation(n_candidates=3, include_per_model=True)
        oracle_matrices, _, _, _ = estimate_oracle_preferences(
            names, evaluation,
            pairwise_delta=0.01,
            use_tail_risk_oracle=False, tail_risk_alpha=0.33,
            use_qre_smoothing=False, qre_temperature_gamma=1.0,
            use_regret_oracle=False,
            use_stability_oracle=True, use_complexity_oracle=True,
            use_robust_oracle=False, use_diversity_oracle=False,
            diversity_oracle_mode="legacy_jaccard",
            diversity_redundancy_weight=0.6,
            diversity_complementarity_weight=0.35,
            performance_oracle_mode="multi_model_oracles",
        )
        self.assertIn("performance_lr_l2", oracle_matrices)
        # fit_tritrust_weights should run successfully with lr_l2 as reference
        weights = fit_tritrust_weights(oracle_matrices)
        self.assertIsInstance(weights, dict)
        self.assertTrue(len(weights) > 0)

    def test_reference_falls_back_to_performance(self):
        """When no per-model oracles, reference falls back to 'performance'."""
        names, evaluation = _make_evaluation(n_candidates=3, include_per_model=False)
        oracle_matrices, _, _, _ = estimate_oracle_preferences(
            names, evaluation,
            pairwise_delta=0.01,
            use_tail_risk_oracle=False, tail_risk_alpha=0.33,
            use_qre_smoothing=False, qre_temperature_gamma=1.0,
            use_regret_oracle=False,
            use_stability_oracle=True, use_complexity_oracle=True,
            use_robust_oracle=False, use_diversity_oracle=False,
            diversity_oracle_mode="legacy_jaccard",
            diversity_redundancy_weight=0.6,
            diversity_complementarity_weight=0.35,
            performance_oracle_mode="single",
        )
        weights = fit_tritrust_weights(oracle_matrices)
        self.assertIsInstance(weights, dict)
        self.assertTrue(len(weights) > 0)

    def test_explicit_reference_override(self):
        """fit_tritrust_weights(reference_oracle=...) uses specified reference."""
        names, evaluation = _make_evaluation(n_candidates=3, include_per_model=True)
        oracle_matrices, _, _, _ = estimate_oracle_preferences(
            names, evaluation,
            pairwise_delta=0.01,
            use_tail_risk_oracle=False, tail_risk_alpha=0.33,
            use_qre_smoothing=False, qre_temperature_gamma=1.0,
            use_regret_oracle=False,
            use_stability_oracle=True, use_complexity_oracle=True,
            use_robust_oracle=False, use_diversity_oracle=False,
            diversity_oracle_mode="legacy_jaccard",
            diversity_redundancy_weight=0.6,
            diversity_complementarity_weight=0.35,
            performance_oracle_mode="multi_model_oracles",
        )
        # Force using aggregate performance as reference
        weights = fit_tritrust_weights(oracle_matrices, reference_oracle="performance")
        self.assertIsInstance(weights, dict)
        self.assertTrue(len(weights) > 0)


class TestWeightNormalization(unittest.TestCase):
    """Per-model oracle weight normalization."""

    def test_oracle_weights_include_per_model(self):
        """When multi_model_oracles, TriTrust weights include per-model keys."""
        names, evaluation = _make_evaluation(n_candidates=3, include_per_model=True)
        oracle_matrices, _, _, _ = estimate_oracle_preferences(
            names, evaluation,
            pairwise_delta=0.01,
            use_tail_risk_oracle=False, tail_risk_alpha=0.33,
            use_qre_smoothing=False, qre_temperature_gamma=1.0,
            use_regret_oracle=False,
            use_stability_oracle=True, use_complexity_oracle=True,
            use_robust_oracle=False, use_diversity_oracle=False,
            diversity_oracle_mode="legacy_jaccard",
            diversity_redundancy_weight=0.6,
            diversity_complementarity_weight=0.35,
            performance_oracle_mode="multi_model_oracles",
        )
        weights = fit_tritrust_weights(oracle_matrices)
        # Per-model oracle keys should appear in weights
        per_model_weight_keys = [k for k in weights if k.startswith("performance_")]
        self.assertGreater(len(per_model_weight_keys), 0)


class TestMissingModelDataFallback(unittest.TestCase):
    """When per-model data is empty, fall back to single oracle only."""

    def test_empty_per_model_data_falls_back(self):
        names, evaluation = _make_evaluation(n_candidates=3, include_per_model=False)
        oracle_matrices, _, _, _ = estimate_oracle_preferences(
            names, evaluation,
            pairwise_delta=0.01,
            use_tail_risk_oracle=False, tail_risk_alpha=0.33,
            use_qre_smoothing=False, qre_temperature_gamma=1.0,
            use_regret_oracle=False,
            use_stability_oracle=False, use_complexity_oracle=False,
            use_robust_oracle=False, use_diversity_oracle=False,
            diversity_oracle_mode="legacy_jaccard",
            diversity_redundancy_weight=0.6,
            diversity_complementarity_weight=0.35,
            performance_oracle_mode="multi_model_oracles",
        )
        # Should only have aggregate performance (no per-model data available)
        self.assertIn("performance", oracle_matrices)
        per_model_keys = [k for k in oracle_matrices if k.startswith("performance_")]
        self.assertEqual(len(per_model_keys), 0)


class TestConfigDataclass(unittest.TestCase):
    """MNPOConfig.performance_oracle_mode defaults and wiring."""

    def test_default_is_single(self):
        cfg = MNPOConfig()
        self.assertEqual(cfg.performance_oracle_mode, "single")

    def test_multi_model_oracles_value(self):
        cfg = MNPOConfig(performance_oracle_mode="multi_model_oracles")
        self.assertEqual(cfg.performance_oracle_mode, "multi_model_oracles")

    def test_config_wires_to_feature_selector(self):
        """FeatureSelectorConfig → FeatureSelector correctly passes performance_oracle_mode."""
        from tabnetics.feature_selection.base import FeatureSelector
        cfg = FeatureSelectorConfig(
            mnpo=MNPOConfig(performance_oracle_mode="multi_model_oracles"),
            evaluation=EvaluationConfig(eval_models_enabled=True),
        )
        fs = FeatureSelector.from_config(cfg)
        self.assertEqual(fs.performance_oracle_mode, "multi_model_oracles")
        self.assertTrue(fs.eval_models_enabled)

    def test_config_default_single_wires(self):
        from tabnetics.feature_selection.base import FeatureSelector
        cfg = FeatureSelectorConfig()
        fs = FeatureSelector.from_config(cfg)
        self.assertEqual(fs.performance_oracle_mode, "single")


class TestEvaluateCandidateLibrary3Tuple(unittest.TestCase):
    """portfolio.evaluate_candidate_library correctly collects per-model fold scores."""

    def test_3tuple_callback_populates_scores_by_model(self):
        """When callback returns 3-tuple, evaluation dict has per-model data."""
        rng = np.random.default_rng(99)
        n, p = 40, 10
        X = rng.normal(size=(n, p))
        y = rng.integers(0, 2, size=n)

        candidates = {
            "dummy_a": {
                "selected_indices": np.array([0, 1, 2]),
                "score_vector": rng.uniform(0, 1, size=p),
                "method_result": {},
            },
            "dummy_b": {
                "selected_indices": np.array([3, 4, 5]),
                "score_vector": rng.uniform(0, 1, size=p),
                "method_result": {},
            },
        }

        def _mock_3tuple(X_tr, y_tr, X_v, y_v):
            score = float(rng.uniform(0.4, 0.9))
            signal = np.zeros(len(y_v), dtype=float)
            per_model = {"lr_l2": score + 0.01, "linear_svc": score - 0.01}
            return score, signal, per_model

        from tabnetics.feature_selection.cv import get_inner_cv_splits

        def _get_splits(X, y):
            return get_inner_cv_splits(X, y, "classification", 42, 3, 2)

        evaluation = evaluate_candidate_library(
            X, y, candidates,
            get_inner_cv_splits_fn=_get_splits,
            fit_and_score_fold_fn=_mock_3tuple,
            augment_training_data_fn=lambda Xtr, ytr: (Xtr, ytr),
            use_robust_oracle=False,
            complexity_use_runtime_penalty=False,
        )
        for name in candidates:
            by_model = evaluation[name].get("performance_scores_by_model", {})
            self.assertIn("lr_l2", by_model)
            self.assertIn("linear_svc", by_model)
            self.assertGreater(len(by_model["lr_l2"]), 0)

    def test_2tuple_callback_empty_per_model(self):
        """When callback returns 2-tuple, per-model data is empty."""
        rng = np.random.default_rng(99)
        n, p = 40, 10
        X = rng.normal(size=(n, p))
        y = rng.integers(0, 2, size=n)

        candidates = {
            "dummy_a": {
                "selected_indices": np.array([0, 1, 2]),
                "score_vector": rng.uniform(0, 1, size=p),
                "method_result": {},
            },
        }

        def _mock_2tuple(X_tr, y_tr, X_v, y_v):
            score = float(rng.uniform(0.4, 0.9))
            signal = np.zeros(len(y_v), dtype=float)
            return score, signal

        from tabnetics.feature_selection.cv import get_inner_cv_splits

        def _get_splits(X, y):
            return get_inner_cv_splits(X, y, "classification", 42, 3, 2)

        evaluation = evaluate_candidate_library(
            X, y, candidates,
            get_inner_cv_splits_fn=_get_splits,
            fit_and_score_fold_fn=_mock_2tuple,
            augment_training_data_fn=lambda Xtr, ytr: (Xtr, ytr),
            use_robust_oracle=False,
            complexity_use_runtime_penalty=False,
        )
        by_model = evaluation["dummy_a"].get("performance_scores_by_model", {})
        self.assertEqual(len(by_model), 0)

    def test_oracle_eval_failure_metadata_propagates(self):
        """Failure diagnostics are captured in evaluation metadata deterministically."""
        rng = np.random.default_rng(17)
        n, p = 30, 8
        X = rng.normal(size=(n, p))
        y = rng.integers(0, 2, size=n)

        candidates = {
            "dummy_a": {
                "selected_indices": np.array([0, 1, 2]),
                "score_vector": rng.uniform(0, 1, size=p),
                "method_result": {},
            },
        }

        call_count = {"n": 0}

        def _mock_3tuple_with_diag(X_tr, y_tr, X_v, y_v):
            call_count["n"] += 1
            if call_count["n"] == 1:
                raise RuntimeError("forced fold exception")
            return 0.5, np.zeros(len(y_v), dtype=float), {
                "lr_l2": 0.6,
                "linear_svc": 0.4,
                "_evaluation_failures_total": 1.0,
                "_evaluation_failures_lr_l2": 1.0,
            }

        from tabnetics.feature_selection.cv import get_inner_cv_splits

        def _get_splits(X, y):
            return get_inner_cv_splits(X, y, "classification", 42, 3, 1)

        evaluation = evaluate_candidate_library(
            X, y, candidates,
            get_inner_cv_splits_fn=_get_splits,
            fit_and_score_fold_fn=_mock_3tuple_with_diag,
            augment_training_data_fn=lambda Xtr, ytr: (Xtr, ytr),
            use_robust_oracle=True,
            complexity_use_runtime_penalty=False,
        )
        failure_meta = evaluation["dummy_a"].get("evaluation_failures", {})
        self.assertGreaterEqual(int(failure_meta.get("fold_exceptions", 0)), 1)
        self.assertGreaterEqual(float(failure_meta.get("model_failures_total", 0.0)), 1.0)
        by_model_fail = failure_meta.get("model_failures_by_model", {})
        self.assertGreaterEqual(float(by_model_fail.get("lr_l2", 0.0)), 1.0)


class TestIntegrationSyntheticFlow(unittest.TestCase):
    """End-to-end: synthetic data → 3-tuple → oracle matrices → TriTrust weights."""

    def test_full_flow(self):
        rng = np.random.default_rng(123)
        n, p = 50, 15
        X = rng.normal(size=(n, p))
        y = rng.integers(0, 2, size=n)

        candidates = {}
        for i in range(3):
            sel = np.sort(rng.choice(p, size=5, replace=False))
            candidates[f"m{i}"] = {
                "selected_indices": sel,
                "score_vector": rng.uniform(0, 1, size=p),
                "method_result": {},
            }

        model_keys = ("lr_l2", "linear_svc", "rf_small")

        def _mock_3tuple(X_tr, y_tr, X_v, y_v):
            score = float(rng.uniform(0.4, 0.9))
            signal = np.zeros(len(y_v), dtype=float)
            per_model = {mk: float(rng.uniform(0.3, 0.95)) for mk in model_keys}
            return score, signal, per_model

        from tabnetics.feature_selection.cv import get_inner_cv_splits

        def _get_splits(X, y):
            return get_inner_cv_splits(X, y, "classification", 42, 3, 2)

        evaluation = evaluate_candidate_library(
            X, y, candidates,
            get_inner_cv_splits_fn=_get_splits,
            fit_and_score_fold_fn=_mock_3tuple,
            augment_training_data_fn=lambda Xtr, ytr: (Xtr, ytr),
            use_robust_oracle=False,
            complexity_use_runtime_penalty=False,
        )

        candidate_names = list(candidates.keys())
        oracle_matrices, oracle_scores, _, _ = estimate_oracle_preferences(
            candidate_names, evaluation,
            pairwise_delta=0.01,
            use_tail_risk_oracle=False, tail_risk_alpha=0.33,
            use_qre_smoothing=False, qre_temperature_gamma=1.0,
            use_regret_oracle=False,
            use_stability_oracle=True, use_complexity_oracle=True,
            use_robust_oracle=False, use_diversity_oracle=False,
            diversity_oracle_mode="legacy_jaccard",
            diversity_redundancy_weight=0.6,
            diversity_complementarity_weight=0.35,
            performance_oracle_mode="multi_model_oracles",
        )

        # Verify structure
        self.assertIn("performance", oracle_matrices)
        for mk in model_keys:
            oracle_name = f"performance_{mk}"
            self.assertIn(oracle_name, oracle_matrices)
            mat = oracle_matrices[oracle_name]
            self.assertEqual(mat.shape, (3, 3))
            np.testing.assert_allclose(np.diag(mat), 0.5)

        # TriTrust should work with multi-model oracles
        weights = fit_tritrust_weights(oracle_matrices)
        self.assertIsInstance(weights, dict)
        self.assertIn("performance_lr_l2", weights)
        self.assertIn("performance_linear_svc", weights)
        self.assertIn("performance_rf_small", weights)


class TestOracleStabilityDiagnostics(unittest.TestCase):
    """Shared oracle stability diagnostics contract tests (T-A003-R2)."""

    def test_diagnostics_deterministic_for_fixed_inputs(self):
        oracle_scores = {
            "performance": np.array([0.80, 0.70, 0.60]),
            "stability": np.array([0.75, 0.65, 0.55]),
            "complexity": np.array([0.20, 0.40, 0.60]),
        }
        d1 = compute_oracle_stability_diagnostics(oracle_scores)
        d2 = compute_oracle_stability_diagnostics(oracle_scores)
        self.assertEqual(d1, d2)
        self.assertIn("pairwise_rank_correlation", d1)
        self.assertEqual(d1["pair_count"], 3)

    def test_estimate_outputs_stability_meta(self):
        names, evaluation = _make_evaluation(n_candidates=4, include_per_model=True)
        _, _, components, pairwise_meta = estimate_oracle_preferences(
            names, evaluation,
            pairwise_delta=0.01,
            use_tail_risk_oracle=False,
            tail_risk_alpha=0.33,
            use_qre_smoothing=False,
            qre_temperature_gamma=1.0,
            use_regret_oracle=False,
            use_stability_oracle=True,
            use_complexity_oracle=True,
            use_robust_oracle=False,
            use_diversity_oracle=False,
            diversity_oracle_mode="legacy_jaccard",
            diversity_redundancy_weight=0.6,
            diversity_complementarity_weight=0.35,
            performance_oracle_mode="multi_model_oracles",
        )
        self.assertIn("oracle_stability", components)
        self.assertIn("oracle_stability", pairwise_meta)
        self.assertIn("mean_rank_correlation", pairwise_meta["oracle_stability"])


if __name__ == "__main__":
    unittest.main()



"""MNPO rank aggregation, wrapper refinement, and legacy consensus.

Standalone functions extracted from FeatureSelector (base.py).
"""

import numpy as np
import scipy.stats as sps
from typing import Dict, Any


# ---------------------------------------------------------------------------
# 1. wrapper_refine_subset_score
# ---------------------------------------------------------------------------
def wrapper_refine_subset_score(X, y, feature_subset,
                                get_inner_cv_splits_fn, fit_and_score_fold_fn):
    """Score one feature subset using the same inner-CV fit/score pathway as MNPO."""
    subset = np.asarray(feature_subset, dtype=int).ravel()
    subset = np.array(sorted(set(int(i) for i in subset if 0 <= int(i) < X.shape[1])), dtype=int)
    if subset.size == 0:
        return float("-inf")

    try:
        splits = get_inner_cv_splits_fn(X, y)
    except Exception as exc:
        return float("-inf")
    if not splits:
        return float("-inf")

    scores = []
    for train_idx, val_idx in splits:
        X_train = X[train_idx][:, subset]
        y_train = y[train_idx]
        X_val = X[val_idx][:, subset]
        y_val = y[val_idx]
        try:
            result = fit_and_score_fold_fn(X_train, y_train, X_val, y_val)
            fold_score = result[0] if isinstance(result, tuple) else result
        except Exception as exc:
            continue
        if np.isfinite(fold_score):
            scores.append(float(fold_score))

    if not scores:
        return float("-inf")
    return float(np.mean(scores))


# ---------------------------------------------------------------------------
# 2. apply_wrapper_refinement
# ---------------------------------------------------------------------------
def apply_wrapper_refinement(
    X, y, vote_ranking, n_final_features,
    *,
    wrapper_refine_enabled,
    wrapper_refine_top_k,
    wrapper_refine_max_add,
    wrapper_refine_min_gain,
    get_inner_cv_splits_fn,
    fit_and_score_fold_fn,
):
    """
    Greedy wrapper refinement over top-ranked features.
    Uses inner-CV score improvements and falls back to vote ranking fill.
    """
    ranking_raw = np.asarray(vote_ranking, dtype=int).ravel()
    seen = set()
    ordered = []
    for idx in ranking_raw:
        idx_i = int(idx)
        if not (0 <= idx_i < X.shape[1]) or idx_i in seen:
            continue
        seen.add(idx_i)
        ordered.append(idx_i)
    ranking = np.asarray(ordered, dtype=int)
    target = int(max(1, min(int(n_final_features), ranking.size)))

    metadata: Dict[str, Any] = {
        "wrapper_refine_enabled": bool(wrapper_refine_enabled),
        "wrapper_refine_applied": False,
        "wrapper_refine_pool_size": 0,
        "wrapper_refine_wrapper_cap": 0,
        "wrapper_refine_min_gain": float(wrapper_refine_min_gain),
        "wrapper_refine_evaluations": 0,
        "wrapper_refine_score_initial": float("nan"),
        "wrapper_refine_score_final": float("nan"),
        "wrapper_refine_stop_reason": "disabled",
    }

    if ranking.size == 0:
        metadata["wrapper_refine_stop_reason"] = "empty_ranking"
        return np.array([], dtype=int), metadata
    if not wrapper_refine_enabled:
        return ranking[:target], metadata

    pool_size = int(min(max(1, wrapper_refine_top_k), ranking.size))
    pool = np.asarray(ranking[:pool_size], dtype=int)
    wrapper_cap = int(min(target, pool.size, max(1, wrapper_refine_max_add)))
    metadata["wrapper_refine_pool_size"] = int(pool_size)
    metadata["wrapper_refine_wrapper_cap"] = int(wrapper_cap)

    selected = [int(pool[0])]
    best_score = wrapper_refine_subset_score(
        X, y, selected, get_inner_cv_splits_fn, fit_and_score_fold_fn,
    )
    metadata["wrapper_refine_score_initial"] = float(best_score if np.isfinite(best_score) else float("nan"))
    evaluations = 1

    stop_reason = "wrapper_cap_reached"
    while len(selected) < wrapper_cap:
        best_feat = None
        best_candidate_score = float("-inf")
        for feat in pool:
            feat_i = int(feat)
            if feat_i in selected:
                continue
            cand = selected + [feat_i]
            score = wrapper_refine_subset_score(
                X, y, cand, get_inner_cv_splits_fn, fit_and_score_fold_fn,
            )
            evaluations += 1
            if score > best_candidate_score:
                best_candidate_score = score
                best_feat = feat_i

        if best_feat is None:
            stop_reason = "no_candidate"
            break

        if np.isfinite(best_score) and np.isfinite(best_candidate_score):
            if best_candidate_score <= (best_score + wrapper_refine_min_gain):
                stop_reason = "min_gain_not_met"
                break

        selected.append(int(best_feat))
        best_score = float(best_candidate_score)

    # Fill remainder from vote ranking to preserve target cardinality.
    selected_set = set(int(i) for i in selected)
    for feat in ranking:
        feat_i = int(feat)
        if feat_i in selected_set:
            continue
        selected.append(feat_i)
        selected_set.add(feat_i)
        if len(selected) >= target:
            break

    refined = np.asarray(selected[:target], dtype=int)
    baseline = np.asarray(ranking[:target], dtype=int)
    metadata["wrapper_refine_applied"] = bool(
        refined.size == baseline.size and np.any(refined != baseline)
    )
    metadata["wrapper_refine_evaluations"] = int(evaluations)
    metadata["wrapper_refine_score_final"] = float(best_score if np.isfinite(best_score) else float("nan"))
    metadata["wrapper_refine_stop_reason"] = str(stop_reason)

    return refined, metadata


# ---------------------------------------------------------------------------
# 3. build_rank_aggregation_candidate
# ---------------------------------------------------------------------------
def build_rank_aggregation_candidate(
    source_candidates, n_target, n_final_features, n_features,
    *,
    rank_aggregation_mode,
    normalize_fn,
):
    """
    Build a synthetic candidate from per-method rankings.
    Supports:
      - Borda: sum of reversed-rank points.
      - RRA: robust rank aggregation score via order-statistic Beta CDF p-values.
    """
    mode = str(rank_aggregation_mode).strip().lower()
    if mode == "none":
        return None
    if len(source_candidates) < 2 or n_features <= 0:
        return None

    rank_cap = int(min(n_features, max(8, int(n_target), int(n_final_features))))
    default_rank = float(rank_cap + 1)
    source_names = list(source_candidates.keys())
    n_sources = int(len(source_names))

    rank_matrix = np.full((n_sources, n_features), default_rank, dtype=float)
    for row, source_name in enumerate(source_names):
        score_vector = np.asarray(
            source_candidates[source_name].get("score_vector", np.zeros(n_features, dtype=float)),
            dtype=float,
        ).ravel()
        if score_vector.size != n_features:
            padded = np.zeros(n_features, dtype=float)
            upto = int(min(n_features, score_vector.size))
            if upto > 0:
                padded[:upto] = score_vector[:upto]
            score_vector = padded
        score_vector = np.nan_to_num(score_vector, nan=0.0, posinf=0.0, neginf=0.0)
        ranked = np.argsort(score_vector)[::-1][:rank_cap]
        if ranked.size > 0:
            rank_matrix[row, ranked] = np.arange(1, ranked.size + 1, dtype=float)

    method_result: Dict[str, Any] = {
        "method": f"rank_aggregate_{mode}",
        "rank_aggregation_mode": mode,
        "rank_aggregation_n_methods": n_sources,
        "rank_aggregation_rank_cap": int(rank_cap),
        "rank_aggregation_source_methods": list(source_names),
    }

    if mode == "borda":
        points = np.maximum(0.0, default_rank - rank_matrix)
        agg_scores = np.sum(points, axis=0)
    else:
        normalized_ranks = np.clip(rank_matrix / default_rank, 1e-12, 1.0)
        sorted_ranks = np.sort(normalized_ranks, axis=0)
        p_values = np.ones(n_features, dtype=float)
        for k in range(n_sources):
            cdf_vals = sps.beta.cdf(sorted_ranks[k], a=float(k + 1), b=float(n_sources - k))
            cdf_vals = np.nan_to_num(cdf_vals, nan=1.0, posinf=1.0, neginf=1.0)
            p_values = np.minimum(p_values, cdf_vals)
        agg_scores = -np.log10(np.clip(p_values, 1e-12, 1.0))
        method_result["rank_aggregation_p_values"] = np.asarray(p_values, dtype=float)

    agg_scores = np.asarray(normalize_fn(agg_scores), dtype=float).ravel()
    agg_scores = np.nan_to_num(agg_scores, nan=0.0, posinf=0.0, neginf=0.0)
    if agg_scores.size != n_features:
        padded = np.zeros(n_features, dtype=float)
        upto = int(min(n_features, agg_scores.size))
        if upto > 0:
            padded[:upto] = agg_scores[:upto]
        agg_scores = padded
    if np.max(agg_scores) <= 0:
        return None

    k_select = int(min(n_features, max(1, int(n_target), int(n_final_features))))
    selected = np.argsort(agg_scores)[::-1][:k_select]
    selected = np.asarray(selected, dtype=int)

    method_result["selected_indices"] = selected
    method_result["scores"] = {int(i): float(agg_scores[i]) for i in selected}
    candidate_name = f"rank_aggregate_{mode}"
    candidate_payload = {
        "selected_indices": selected,
        "score_vector": agg_scores,
        "runtime_sec": 0.0,
        "method_result": method_result,
    }
    return candidate_name, candidate_payload



from collections import Counter, defaultdict

from tabnetics.datasets.benchmark_catalog import DATASET_SETS
from tabnetics.validation.generate_plan import (
    VALIDATION_SEEDS,
    _load_hf_manifest_metadata,
    _balanced_shard_assign_validation10_pairs,
    build_jobs_validation10,
)


def _profile_from_job_id(job_id: str) -> str:
    parts = str(job_id).split("/")
    assert len(parts) == 3
    return parts[1]


def test_build_jobs_validation10_has_expected_job_topology(tmp_path):
    jobs = build_jobs_validation10(dataset_shards=6, val9_root=tmp_path)
    assert len(jobs) == 12  # 2 profiles x 6 dataset shards

    profile_counts = Counter(_profile_from_job_id(j.job_id) for j in jobs)
    assert profile_counts == {
        "simple_all_stages": 6,
        "mnpo_all_stages": 6,
    }

    for job in jobs:
        assert job.kind == "run_df_fs_sota_benchmark"
        assert list(job.params.get("seeds") or []) == list(VALIDATION_SEEDS)


def test_build_jobs_validation10_uses_extended_validation_catalog(tmp_path):
    jobs = build_jobs_validation10(dataset_shards=6, val9_root=tmp_path)
    validation_all = {str(ds) for ds in DATASET_SETS.get("validation_all", [])}
    hf_ids = {str(ds) for ds in _load_hf_manifest_metadata().keys()}
    expected = validation_all & hf_ids
    assert len(expected) >= 68
    assert any(str(ds).startswith("rv_") for ds in expected)

    by_profile = defaultdict(set)
    for job in jobs:
        profile = _profile_from_job_id(job.job_id)
        datasets = {str(ds) for ds in (job.params.get("datasets") or [])}
        by_profile[profile].update(datasets)

    for profile in ("simple_all_stages", "mnpo_all_stages"):
        assert by_profile[profile] == expected


def test_build_jobs_validation10_profiles_encode_stage_modes(tmp_path):
    jobs = build_jobs_validation10(dataset_shards=6, val9_root=tmp_path)
    by_profile = defaultdict(list)
    for job in jobs:
        by_profile[_profile_from_job_id(job.job_id)].append(job)

    def _args(profile: str):
        return list(by_profile[profile][0].params.get("extra_args") or [])

    simple_args = _args("simple_all_stages")
    assert "--dist-criterion" in simple_args
    assert simple_args[simple_args.index("--dist-criterion") + 1] == "simple"
    assert "--classifier-selection-mode" in simple_args
    assert simple_args[simple_args.index("--classifier-selection-mode") + 1] == "legacy"
    assert "--classification-backend" in simple_args
    assert simple_args[simple_args.index("--classification-backend") + 1] == "sklearn"

    mnpo_args = _args("mnpo_all_stages")
    assert "--dist-criterion" in mnpo_args
    assert mnpo_args[mnpo_args.index("--dist-criterion") + 1] == "mnpo_oracle"
    assert "--classifier-selection-mode" in mnpo_args
    assert mnpo_args[mnpo_args.index("--classifier-selection-mode") + 1] == "mnpo_hybrid"
    assert "--enable-fs-rashomon" in mnpo_args
    assert "--df-mnpo-include-preq" in mnpo_args
    assert "--fs-use-conformal-uq" in mnpo_args


def test_validation10_paired_sharding_keeps_profiles_together(tmp_path):
    jobs = build_jobs_validation10(dataset_shards=6, val9_root=tmp_path)
    shards = _balanced_shard_assign_validation10_pairs(jobs, num_shards=6)
    assert set(shards.keys()) == {1, 2, 3, 4, 5, 6}
    for _, job_ids in shards.items():
        assert len(job_ids) == 2
        profiles = {_profile_from_job_id(jid) for jid in job_ids}
        assert profiles == {"simple_all_stages", "mnpo_all_stages"}



import numpy as np
from sklearn.datasets import make_classification

from tabnetics.feature_selection.methods.filter import wmw_auc_selection


def _make_multiclass(seed: int = 0):
    X, y = make_classification(
        n_samples=120,
        n_features=24,
        n_informative=10,
        n_redundant=2,
        n_repeated=0,
        n_classes=4,
        n_clusters_per_class=1,
        random_state=seed,
    )
    return X, y


def test_wmw_auc_multiclass_returns_expected_schema():
    X, y = _make_multiclass(seed=11)
    results, all_scores = wmw_auc_selection(X, y, n_target_features=8, problem_type="classification")
    assert "selected_indices" in results
    assert "scores" in results
    assert "all_scores" in results
    assert "m_metric_scores" in results
    assert "n_classes" in results
    assert results["n_classes"] == 4
    assert len(results["selected_indices"]) == 8
    assert len(all_scores) == X.shape[1]


def test_wmw_auc_multiclass_is_label_permutation_invariant():
    X, y = _make_multiclass(seed=12)
    results_a, _ = wmw_auc_selection(X, y, n_target_features=10, problem_type="classification")
    # Permute labels: 0->2, 1->3, 2->1, 3->0
    mapping = {0: 2, 1: 3, 2: 1, 3: 0}
    y_perm = np.vectorize(mapping.get)(y)
    results_b, _ = wmw_auc_selection(X, y_perm, n_target_features=10, problem_type="classification")
    np.testing.assert_allclose(
        np.asarray(results_a["all_scores"], dtype=float),
        np.asarray(results_b["all_scores"], dtype=float),
        atol=1e-12,
    )


def test_wmw_auc_multiclass_constant_feature_gets_zero_score():
    X, y = _make_multiclass(seed=13)
    X = np.asarray(X, dtype=float)
    X[:, 0] = 1.2345
    results, _ = wmw_auc_selection(X, y, n_target_features=6, problem_type="classification")
    assert float(results["all_scores"][0]) == 0.0


def test_wmw_auc_multiclass_rejects_non_classification_problem_type():
    X, y = _make_multiclass(seed=14)
    results, all_scores = wmw_auc_selection(X, y, n_target_features=6, problem_type="regression")
    assert results == {}
    assert all_scores == {}



"""Core numerical, compatibility, and runtime helpers."""

from .mnpo import *  # noqa: F401,F403
from .compat import *  # noqa: F401,F403
from .errors import *  # noqa: F401,F403
from .paths import find_repo_root
from .runtime import configure_runtime_environment, get_sklearn_n_jobs, set_sklearn_n_jobs

__all__ = [
    "configure_runtime_environment",
    "find_repo_root",
    "get_sklearn_n_jobs",
    "set_sklearn_n_jobs",
]



"""Feature selection method implementations.

Submodules
----------
screening : Tier 2 interaction-aware screening (STIR/ReliefF).
filter    : Univariate filter methods (MI, F-test, WMW-AUC, …).
embedded  : Embedded methods (Lasso, SVM-L1, RF, IPSS, …).
pairwise  : Pairwise methods (k-TSP).
wrapper   : Wrapper/iterative-pruning methods.
knockoff  : Copula knockoff selection (DTDCKe).
multiclass: Multiclass-specific methods (OVA, ECOC, NSC, …).
hsic      : HSIC Lasso.
sdr       : SIR/SAVE/PFC sufficient-dimension reduction selectors.
"""

__all__ = [
    "screening",
    "filter",
    "embedded",
    "pairwise",
    "wrapper",
    "knockoff",
    "multiclass",
    "hsic",
    "sdr",
]



"""Tier 2 interaction-aware screening methods (T-004).

This module implements screening methods that capture feature interactions,
complementing the Tier 1 univariate/bivariate prefilter in ``prefilter.py``.

Architecture (ArchitectureRefactor.md §14 RISK-2):
- Tier 1 (prefilter.py): cheap, always-on, univariate (MI, F-test)
- Tier 2 (this module): expensive, opt-in, interaction-aware (STIR/ReliefF)

Execution flow:
    Tier 1 prefilter → Tier 2 screening → method selection
"""

from __future__ import annotations

import logging
from typing import Optional

import numpy as np
from sklearn.feature_selection import f_classif

logger = logging.getLogger(__name__)

try:
    from ..prefilter import pvalues_to_evalues, ebh_support
except Exception as exc:
    try:
        from tabnetics.feature_selection.prefilter import pvalues_to_evalues, ebh_support
    except Exception as exc:
        def pvalues_to_evalues(p_values):  # type: ignore[misc]
            p = np.asarray(p_values, dtype=float).ravel()
            p = np.clip(np.nan_to_num(p, nan=1.0), 1e-12, 1.0)
            return np.asarray(1.0 / p, dtype=float)

        def ebh_support(e_values, alpha):  # type: ignore[misc]
            e_vals = np.asarray(e_values, dtype=float).ravel()
            if e_vals.size == 0:
                return np.array([], dtype=int)
            order = np.argsort(e_vals)[::-1]
            e_sorted = e_vals[order]
            p = int(e_vals.size)
            thresh = np.maximum((np.arange(1, p + 1, dtype=float) * float(alpha)) / float(p), 1e-12)
            valid = np.where(e_sorted >= (1.0 / thresh))[0]
            if valid.size == 0:
                return np.array([], dtype=int)
            return np.sort(order[: int(valid.max()) + 1])


# ---------------------------------------------------------------------------
# STIR / ReliefF scoring
# ---------------------------------------------------------------------------

def _nearest_neighbors(
    X: np.ndarray,
    idx: int,
    y: np.ndarray,
    same_class: bool,
    k: int,
) -> np.ndarray:
    """Find *k* nearest neighbours of ``X[idx]`` from same/different class.

    Parameters
    ----------
    X : (n, p) standardised feature matrix
    idx : row index of the query instance
    y : (n,) label array
    same_class : if True, search within the same class; else different class
    k : number of neighbours to return

    Returns
    -------
    indices : 1-d int array of neighbour row indices (may be shorter than *k*
        if fewer candidates are available)
    """
    label = y[idx]
    if same_class:
        mask = (y == label)
    else:
        mask = (y != label)
    mask[idx] = False  # exclude the query itself

    candidate_idx = np.flatnonzero(mask)
    if candidate_idx.size == 0:
        return np.array([], dtype=int)

    # Euclidean distances to candidates
    diffs = X[candidate_idx] - X[idx]
    dists = np.einsum("ij,ij->i", diffs, diffs)  # squared L2
    k_eff = min(k, candidate_idx.size)
    # argpartition is O(n) vs O(n log n) for argsort
    if k_eff < candidate_idx.size:
        part = np.argpartition(dists, k_eff - 1)[:k_eff]
    else:
        part = np.arange(candidate_idx.size)
    return candidate_idx[part]


def compute_stir_scores(
    X: np.ndarray,
    y: np.ndarray,
    n_neighbors: int = 10,
    n_iter: int = 50,
    random_state: Optional[int] = None,
) -> np.ndarray:
    """Compute STIR/ReliefF-style feature importance scores.

    For each iteration:
    1. Sample a random instance
    2. Find *k* nearest neighbours of same class ("hits")
    3. Find *k* nearest neighbours of different class ("misses")
    4. Update feature weights: features that differ more from misses
       but agree with hits get higher scores

    For **multiclass** problems, miss contributions are weighted by
    the prior probability of each foreign class (standard ReliefF
    multi-class extension).

    Parameters
    ----------
    X : (n_samples, n_features)
    y : (n_samples,)
    n_neighbors : number of nearest neighbours per query
    n_iter : number of sampling iterations (hard cap, not wall-clock)
    random_state : seed for deterministic sampling

    Returns
    -------
    scores : (n_features,) array; higher = more relevant.
        All values are finite (NaN/Inf replaced by 0).
    """
    X = np.asarray(X, dtype=np.float64)
    y = np.asarray(y).ravel()

    n_samples, n_features = X.shape

    # --- Edge cases --------------------------------------------------------
    if n_features == 0:
        return np.zeros(0, dtype=np.float64)

    classes, class_counts = np.unique(y, return_counts=True)
    n_classes = classes.size

    if n_classes < 2:
        # Single class → no discrimination possible
        logger.warning("compute_stir_scores: single class detected; returning zeros")
        return np.zeros(n_features, dtype=np.float64)

    if n_samples < 2:
        logger.warning("compute_stir_scores: n_samples < 2; returning zeros")
        return np.zeros(n_features, dtype=np.float64)

    # --- Standardise features (zero-mean, unit-var) ------------------------
    col_std = np.std(X, axis=0, ddof=0)
    constant_mask = col_std < 1e-12
    col_std[constant_mask] = 1.0  # avoid divide-by-zero; constant cols get 0 weight
    X_std = (X - np.mean(X, axis=0)) / col_std

    # --- Class priors for multiclass weighting of misses -------------------
    priors = class_counts.astype(np.float64) / n_samples
    class_to_prior = dict(zip(classes, priors))

    # --- Main loop ---------------------------------------------------------
    rng = np.random.RandomState(random_state)
    weights = np.zeros(n_features, dtype=np.float64)

    effective_iter = min(n_iter, n_samples)  # can't sample more than n

    for _ in range(effective_iter):
        idx = rng.randint(0, n_samples)
        label_i = y[idx]

        # --- Hits (same class) ---
        hit_idx = _nearest_neighbors(X_std, idx, y, same_class=True, k=n_neighbors)
        if hit_idx.size > 0:
            hit_diffs = np.abs(X_std[idx] - X_std[hit_idx])
            mean_hit_diff = np.mean(hit_diffs, axis=0)
            weights -= mean_hit_diff

        # --- Misses (different classes) ---
        for c in classes:
            if c == label_i:
                continue
            # Prior weight = P(class) / (1 - P(label_i))
            p_c = class_to_prior[c]
            p_label = class_to_prior[label_i]
            denom = 1.0 - p_label
            if denom < 1e-12:
                continue
            miss_weight = p_c / denom

            # Neighbours of class c
            class_mask = (y == c)
            class_candidates = np.flatnonzero(class_mask)
            if class_candidates.size == 0:
                continue

            k_eff = min(n_neighbors, class_candidates.size)
            diffs_c = X_std[class_candidates] - X_std[idx]
            dists_c = np.einsum("ij,ij->i", diffs_c, diffs_c)
            if k_eff < class_candidates.size:
                part = np.argpartition(dists_c, k_eff - 1)[:k_eff]
            else:
                part = np.arange(class_candidates.size)
            miss_nn = class_candidates[part]

            miss_diffs = np.abs(X_std[idx] - X_std[miss_nn])
            mean_miss_diff = np.mean(miss_diffs, axis=0)
            weights += miss_weight * mean_miss_diff

    # Normalise by number of iterations
    if effective_iter > 0:
        weights /= effective_iter

    # Zero-out constant features
    weights[constant_mask] = 0.0

    # Sanitise
    weights = np.nan_to_num(weights, nan=0.0, posinf=0.0, neginf=0.0)
    return weights


# ---------------------------------------------------------------------------
# Screening wrapper
# ---------------------------------------------------------------------------
def _screen_features_evalue(
    X: np.ndarray,
    y: np.ndarray,
    *,
    alpha: float,
    min_features: int,
) -> Optional[np.ndarray]:
    """Tier 2 e-value screening (p→e conversion + e-BH support)."""
    n_samples, n_features = X.shape
    if n_features == 0 or n_samples < 2:
        return np.array([], dtype=int)
    y_arr = np.asarray(y).ravel()
    if np.unique(y_arr).size < 2:
        return None

    try:
        _, pvals = f_classif(X, y_arr)
        pvals = np.asarray(pvals, dtype=float).ravel()
    except Exception as exc:
        logger.exception("E-value screening: failed to compute p-values")
        return None
    pvals = np.nan_to_num(pvals, nan=1.0, posinf=1.0, neginf=1.0)
    e_vals = pvalues_to_evalues(pvals)
    support = ebh_support(e_vals, alpha=float(alpha))
    support = np.asarray(
        sorted(set(int(i) for i in np.asarray(support, dtype=int).tolist() if 0 <= int(i) < n_features)),
        dtype=int,
    )
    if support.size < int(max(1, min_features)):
        fallback_k = int(min(n_features, max(1, min_features)))
        fallback = np.argsort(e_vals)[::-1][:fallback_k]
        support = np.asarray(sorted(set(int(i) for i in fallback.tolist())), dtype=int)
    if support.size >= n_features:
        return None
    return support


def screen_features(
    X: np.ndarray,
    y: np.ndarray,
    *,
    enabled: bool = False,
    method: str = "none",
    pool_cap: int = 2000,
    stir_n_neighbors: int = 10,
    stir_n_iter: int = 50,
    stir_keep_fraction: float = 0.5,
    stir_min_features: int = 20,
    evalue_alpha: float = 0.2,
    evalue_min_features: int = 20,
    random_state: Optional[int] = None,
) -> Optional[np.ndarray]:
    """Apply Tier 2 STIR screening to reduce the feature pool.

    Parameters
    ----------
    X : (n_samples, n_features) — already prefiltered by Tier 1
    y : (n_samples,) — labels
    enabled : screening toggle
    method : screening method ("stir" | "none")
    pool_cap : safety cap on features to screen
    stir_n_neighbors, stir_n_iter : STIR hyper-parameters
    stir_keep_fraction : fraction of features to retain (0, 1]
    stir_min_features : never drop below this many features
    random_state : seed

    Returns
    -------
    selected_indices : 1-d int array of column indices to keep
        (relative to *X*), or ``None`` if screening is disabled /
        not applicable.
    """
    if not enabled or method == "none":
        return None

    n_samples, n_features = X.shape

    if n_features == 0:
        return np.array([], dtype=int)

    # --- Pool cap: subsample columns if too wide -------------------------
    if n_features > pool_cap:
        logger.info(
            "Tier 2 screening: pool_cap=%d reached (p=%d); subsampling columns",
            pool_cap, n_features,
        )
        rng = np.random.RandomState(random_state)
        keep = rng.choice(n_features, size=pool_cap, replace=False)
        keep.sort()
        X_screen = X[:, keep]
        index_map = keep
    else:
        X_screen = X
        index_map = np.arange(n_features)

    method_norm = str(method or "none").strip().lower()
    if method_norm not in {"stir", "evalue"}:
        logger.warning("Unknown screening method %r; skipping", method)
        return None

    if method_norm == "evalue":
        selected_screen = _screen_features_evalue(
            X_screen,
            y,
            alpha=float(evalue_alpha),
            min_features=int(evalue_min_features),
        )
        if selected_screen is None:
            return None
        selected_screen = np.asarray(selected_screen, dtype=int)
        selected = index_map[selected_screen]
        logger.info(
            "Tier 2 e-value screening: %d → %d features retained (alpha=%.3f, min=%d)",
            n_features,
            selected.size,
            float(evalue_alpha),
            int(evalue_min_features),
        )
        return np.asarray(selected, dtype=int)

    # method_norm == "stir"
    try:
        scores = compute_stir_scores(
            X_screen, y,
            n_neighbors=stir_n_neighbors,
            n_iter=stir_n_iter,
            random_state=random_state,
        )
    except Exception as exc:
        logger.exception("STIR screening failed; falling back to no screening")
        return None

    n_keep = max(
        int(np.ceil(stir_keep_fraction * len(scores))),
        min(stir_min_features, len(scores)),
    )
    n_keep = min(n_keep, len(scores))
    if n_keep >= len(scores):
        return None

    top_idx = np.argsort(scores)[::-1][:n_keep]
    top_idx.sort()
    selected = index_map[top_idx]
    logger.info(
        "Tier 2 STIR screening: %d → %d features retained "
        "(keep_fraction=%.2f, min_features=%d)",
        n_features, selected.size, stir_keep_fraction, stir_min_features,
    )
    return np.asarray(selected, dtype=int)


def screen_features_stir(
    X: np.ndarray,
    y: np.ndarray,
    *,
    enabled: bool = False,
    method: str = "none",
    pool_cap: int = 2000,
    stir_n_neighbors: int = 10,
    stir_n_iter: int = 50,
    stir_keep_fraction: float = 0.5,
    stir_min_features: int = 20,
    random_state: Optional[int] = None,
) -> Optional[np.ndarray]:
    """Backward-compatible wrapper for prior callers/tests."""
    return screen_features(
        X,
        y,
        enabled=enabled,
        method=method,
        pool_cap=pool_cap,
        stir_n_neighbors=stir_n_neighbors,
        stir_n_iter=stir_n_iter,
        stir_keep_fraction=stir_keep_fraction,
        stir_min_features=stir_min_features,
        random_state=random_state,
    )



import numpy as np
import scipy.stats as sps
import warnings
from typing import Dict, Tuple, List, Any

__all__ = [
    "BootstrapGOFSelector",
    "select_best_distribution_bootstrap",
    "select_best_distribution_bic",
    "auto_select_distribution",
]


class BootstrapGOFSelector:
    """Monte-Carlo goodness-of-fit selector tailored to *small* samples.

    This class recalibrates Kolmogorov--Smirnov or Cram\u00E9r-von Mises statistics
    via parametric bootstrapping.  For every candidate distribution we:
      1. Fit parameters (by MLE or SciPy default `fit`).
      2. Compute the observed test statistic.
      3. Simulate *n_boot* synthetic samples **of the same size** as the data,
         re-fitting the parameters for *each* replicate to account for
         estimation uncertainty.
      4. Estimate the p-value as the proportion of bootstrapped statistics that
         are **at least as extreme** as the observed statistic.

    An empirical p-value obtained this way is valid for tiny *n* because it
    respects parameter uncertainty and small-sample variability, unlike the
    asymptotic tables used by vanilla KS/CvM in SciPy.
    """

    SUPPORTED_STATS = {"ks", "cvm"}

    def __init__(self, statistic: str = "cvm", n_boot: int = 2000, random_state: Any = None):
        if statistic not in self.SUPPORTED_STATS:
            raise ValueError(f"statistic must be one of {self.SUPPORTED_STATS}")
        self.statistic = statistic
        self.n_boot = int(n_boot)
        self.rng = np.random.default_rng(random_state)

    # ---------------------------------------------------------------------
    # Public helpers
    # ---------------------------------------------------------------------

    def p_value(self, data: np.ndarray, dist: sps.rv_continuous, params: Tuple) -> Tuple[float, float]:
        """Return (p-value, observed_statistic)."""
        obs_stat = self._compute_stat(data, dist, params)
        boot_stats = np.empty(self.n_boot)

        for i in range(self.n_boot):
            sim = dist.rvs(*params[:-2], loc=params[-2], scale=params[-1], size=data.size, random_state=self.rng)
            # Re-fit on each bootstrap sample to mimic real workflow
            try:
                sim_params = dist.fit(sim)
            except Exception as exc:
                # If re-fit fails, fall back to original params – conservative.
                sim_params = params
            boot_stats[i] = self._compute_stat(sim, dist, sim_params)

        p_val = (boot_stats >= obs_stat).mean()
        return p_val, obs_stat

    # ------------------------------------------------------------------
    # Internals
    # ------------------------------------------------------------------

    def _compute_stat(self, sample: np.ndarray, dist: sps.rv_continuous, params: Tuple) -> float:
        if self.statistic == "ks":
            return sps.kstest(sample, dist.cdf, args=params).statistic
        # default cvm
        return sps.cramervonmises(sample, dist.cdf, args=params).statistic


# -------------------------------------------------------------------------
#  Bootstrap-based selection
# -------------------------------------------------------------------------

def _default_dist_dict() -> Dict[str, sps.rv_continuous]:
    return {
        "norm": sps.norm,
        "expon": sps.expon,
        "weibull_min": sps.weibull_min,
        "gamma": sps.gamma,
        "lognorm": sps.lognorm,
        "beta": sps.beta,
        "uniform": sps.uniform,
    }


def select_best_distribution_bootstrap(
    data: np.ndarray,
    distributions: Dict[str, sps.rv_continuous] | None = None,
    statistic: str = "cvm",
    n_boot: int = 2000,
    random_state: Any = None,
) -> Tuple[str, List[Dict[str, Any]]]:
    """Return (best_dist_name, sorted_result_list).

    Each result is a dict:  {"name", "p_value", "stat", "params"} sorted by
    descending p-value.  Best distribution is simply the first entry.
    """
    if distributions is None:
        distributions = _default_dist_dict()

    selector = BootstrapGOFSelector(statistic, n_boot, random_state)
    results: List[Dict[str, Any]] = []

    for name, dist in distributions.items():
        try:
            params = dist.fit(data)
            p_val, obs_stat = selector.p_value(data, dist, params)
            results.append({"name": name, "p_value": p_val, "stat": obs_stat, "params": params})
        except Exception as exc:
            warnings.warn(f"{name}: fit failed – {exc}")
            continue

    if not results:
        raise RuntimeError("No distribution could be fitted; cannot select best.")

    # Higher p-value ⇒ better fit
    results.sort(key=lambda d: d["p_value"], reverse=True)
    return results[0]["name"], results


# -------------------------------------------------------------------------
#  BIC — Bayesian Information Criterion (quick Bayesian proxy)
# -------------------------------------------------------------------------

def _bic(loglik: float, n: int, k: int) -> float:
    return -2.0 * loglik + k * np.log(n)


def select_best_distribution_bic(
    data: np.ndarray,
    distributions: Dict[str, sps.rv_continuous] | None = None,
) -> Tuple[str, List[Dict[str, Any]]]:
    """Return (best_dist_name, sorted_result_list) using BIC."""
    if distributions is None:
        distributions = _default_dist_dict()

    n = data.size
    results: List[Dict[str, Any]] = []

    for name, dist in distributions.items():
        try:
            params = dist.fit(data)
            loglik = dist.logpdf(data, *params).sum()
            bic_val = _bic(loglik, n, len(params))
            results.append({"name": name, "bic": bic_val, "params": params})
        except Exception as exc:
            warnings.warn(f"{name}: fit failed – {exc}")
            continue

    if not results:
        raise RuntimeError("No distribution could be fitted; cannot select best.")

    # Lower BIC ⇒ better (more parsimonious) fit
    results.sort(key=lambda d: d["bic"])
    return results[0]["name"], results


# -------------------------------------------------------------------------
#  Auto strategy: small-n ⇒ bootstrap, else fall back
# -------------------------------------------------------------------------

def auto_select_distribution(
    data: np.ndarray,
    *,
    small_sample_threshold: int = 30,
    statistic: str = "cvm",
    n_boot: int = 2000,
    random_state: Any = None,
) -> Tuple[str, List[Dict[str, Any]]]:
    """Dispatch to the most appropriate selection strategy.

    By default:
      • If n ≤ *small_sample_threshold* (30), use bootstrapped CvM.
      • Otherwise, pick the distribution with the lowest BIC.

    This heuristic blends a computationally intensive but accurate method for
    tiny datasets with a fast, information-criterion method for larger samples.
    """
    n = data.size
    if n <= small_sample_threshold:
        return select_best_distribution_bootstrap(data, statistic=statistic, n_boot=n_boot, random_state=random_state)
    return select_best_distribution_bic(data)




import numpy as np
from sklearn.datasets import make_classification

from tabnetics.feature_selection.base import FeatureSelector
from tabnetics.feature_selection.mnpo.portfolio import (
    extract_portfolio,
    resolve_adaptive_portfolio_size,
)


def _make_data(seed: int = 7):
    X, y = make_classification(
        n_samples=72,
        n_features=80,
        n_informative=12,
        n_redundant=4,
        n_classes=3,
        n_clusters_per_class=1,
        random_state=seed,
    )
    return X.astype(float), y.astype(int)


def test_resolve_adaptive_portfolio_size_disabled_returns_requested_bound():
    w = np.array([0.6, 0.3, 0.1], dtype=float)
    k, meta = resolve_adaptive_portfolio_size(
        w,
        portfolio_size=4,
        adaptive_enabled=False,
        adaptive_size_min=2,
        adaptive_size_max=5,
    )
    assert int(k) == 3
    assert str(meta["reason"]) == "disabled"


def test_resolve_adaptive_portfolio_size_head_heavy_prefers_small_k():
    w = np.array([0.78, 0.12, 0.05, 0.03, 0.02], dtype=float)
    k, meta = resolve_adaptive_portfolio_size(
        w,
        portfolio_size=5,
        adaptive_enabled=True,
        adaptive_size_min=2,
        adaptive_size_max=5,
    )
    assert 2 <= int(k) <= 4
    assert str(meta["reason"]) == "ok"
    assert int(meta["k_mass"]) >= 1
    assert int(meta["k_elbow"]) >= 1


def test_resolve_adaptive_portfolio_size_variance_penalty_shrinks_k():
    w = np.array([0.85, 0.10, 0.03, 0.01, 0.01], dtype=float)
    k_base, _ = resolve_adaptive_portfolio_size(
        w,
        portfolio_size=5,
        adaptive_enabled=True,
        adaptive_size_min=2,
        adaptive_size_max=5,
        adaptive_sizing_variance_penalty=False,
    )
    k_pen, meta_pen = resolve_adaptive_portfolio_size(
        w,
        portfolio_size=5,
        adaptive_enabled=True,
        adaptive_size_min=2,
        adaptive_size_max=5,
        adaptive_sizing_variance_penalty=True,
        adaptive_sizing_variance_penalty_strength=1.0,
    )
    assert int(k_pen) <= int(k_base)
    assert bool(meta_pen["variance_penalty_enabled"]) is True


def test_resolve_adaptive_portfolio_size_missing_bounds_falls_back():
    w = np.array([0.4, 0.3, 0.2, 0.1], dtype=float)
    k, meta = resolve_adaptive_portfolio_size(
        w,
        portfolio_size=3,
        adaptive_enabled=True,
        adaptive_size_min=None,
        adaptive_size_max=5,
    )
    assert int(k) == 3
    assert str(meta["reason"]) == "missing_bounds"


def test_feature_selector_emits_adaptive_portfolio_metadata():
    X, y = _make_data()
    sel = FeatureSelector(
        random_state=11,
        n_bootstrap_iterations=1,
        inner_cv_splits=3,
        inner_cv_repeats=1,
        mirror_descent_steps=40,
        enabled_methods={"mutual_information", "anova_f", "linear_svm"},
        selection_strategy="mnpo_portfolio",
        portfolio_size=5,
        adaptive_portfolio_sizing_enabled=True,
        adaptive_size_min=3,
        adaptive_size_max=5,
    )

    _, result = sel.fit_transform(X, y, n_final_features=12, return_result_object=True)
    mnpo = dict(result.method_results.get("mnpo_portfolio", {}) or {})
    meta = dict(mnpo.get("adaptive_portfolio_sizing", {}) or {})

    assert bool(meta.get("adaptive_enabled", False)) is True
    assert int(mnpo.get("portfolio_size_requested", 0)) == 5
    assert 3 <= int(mnpo.get("portfolio_size_effective", 0)) <= 5


def _entry(indices, signal):
    return {
        "selected_indices": np.asarray(indices, dtype=int),
        "prediction_signal": np.asarray(signal, dtype=float),
    }


def test_extract_portfolio_blocks_near_clone_by_overlap_alone():
    names = ["m0", "m1", "m2"]
    weights = np.asarray([0.90, 0.80, 0.70], dtype=float)
    evaluation = {
        "m0": _entry(range(10), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
        "m1": _entry([0, 1, 2, 3, 4, 5, 6, 7, 8, 20], [0, 1, 0, 1, 0, 1, 0, 1, 0, 1]),
        "m2": _entry(range(30, 40), [0, 0, 1, 1, 0, 0, 1, 1, 0, 0]),
    }
    selected = extract_portfolio(
        names, weights, evaluation, portfolio_size=2, use_diversity_oracle=True
    )
    assert selected == [0, 2]


def test_extract_portfolio_blocks_near_clone_by_correlation_alone():
    names = ["m0", "m1", "m2"]
    weights = np.asarray([0.90, 0.80, 0.70], dtype=float)
    base_signal = np.asarray([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.8, 1.0, 1.2, 1.4], dtype=float)
    evaluation = {
        "m0": _entry(range(10), base_signal),
        "m1": _entry(range(20, 30), 2.0 * base_signal + 1.0),
        "m2": _entry(range(40, 50), [1, 0, 1, 0, 0, 1, 0, 1, 0, 1]),
    }
    selected = extract_portfolio(
        names, weights, evaluation, portfolio_size=2, use_diversity_oracle=True
    )
    assert selected == [0, 2]


def test_extract_portfolio_keeps_diverse_pair():
    names = ["m0", "m1"]
    weights = np.asarray([0.90, 0.80], dtype=float)
    evaluation = {
        "m0": _entry(range(10), [0, 1, 1, 0, 1, 0, 1, 0, 1, 0]),
        "m1": _entry(range(15, 25), [1, 0, 0, 1, 0, 1, 0, 1, 0, 1]),
    }
    selected = extract_portfolio(
        names, weights, evaluation, portfolio_size=2, use_diversity_oracle=True
    )
    assert selected == [0, 1]


def test_extract_portfolio_keeps_old_and_block_cases_blocked():
    names = ["m0", "m1", "m2"]
    weights = np.asarray([0.90, 0.80, 0.70], dtype=float)
    evaluation = {
        "m0": _entry(range(10), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
        "m1": _entry(range(10), [0.01, 1.01, 2.01, 3.01, 4.01, 5.01, 6.01, 7.01, 8.01, 9.01]),
        "m2": _entry(range(30, 40), [0, 1, 0, 1, 0, 1, 0, 1, 0, 1]),
    }
    selected = extract_portfolio(
        names, weights, evaluation, portfolio_size=2, use_diversity_oracle=True
    )
    assert selected == [0, 2]



"""Tests for knockpy/GRIP2 benchmark selector (VAL12_Suggestions §5.2)."""

import numpy as np
import pytest
from unittest.mock import patch, MagicMock

from tabnetics.feature_selection.methods.knockoff_benchmark import (
    KnockoffBenchmarkSelector,
    knockoff_benchmark_selection,
    KNOCKPY_AVAILABLE,
)


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _make_binary_data(n=100, p=20, n_informative=5, seed=42):
    """Create synthetic data with known informative features."""
    rng = np.random.RandomState(seed)
    X = rng.randn(n, p)
    beta = np.zeros(p)
    beta[:n_informative] = rng.uniform(1.0, 3.0, size=n_informative)
    logits = X @ beta
    prob = 1.0 / (1.0 + np.exp(-logits))
    y = (prob > 0.5).astype(float)
    return X, y


# ---------------------------------------------------------------------------
# Fallback tests (no knockpy needed)
# ---------------------------------------------------------------------------


class TestKnockoffFallback:
    """Verify behavior when knockpy is unavailable."""

    def test_unavailable_returns_empty(self):
        X, y = _make_binary_data()
        with patch(
            "tabnetics.feature_selection.methods.knockoff_benchmark.KNOCKPY_AVAILABLE",
            False,
        ):
            selector = KnockoffBenchmarkSelector(fdr=0.10, random_state=42)
            with pytest.warns(RuntimeWarning, match="knockpy not installed"):
                result, meta = selector.select(X, y)
        assert result["selected_count"] == 0
        assert result["selected_indices"].size == 0
        assert meta["knockoff_backend"] == "unavailable"

    def test_empty_data(self):
        selector = KnockoffBenchmarkSelector(random_state=42)
        with patch(
            "tabnetics.feature_selection.methods.knockoff_benchmark.KNOCKPY_AVAILABLE",
            True,
        ):
            result, meta = selector.select(np.zeros((1, 0)), np.array([0]))
        assert result["selected_count"] == 0

    def test_single_sample(self):
        selector = KnockoffBenchmarkSelector(random_state=42)
        with patch(
            "tabnetics.feature_selection.methods.knockoff_benchmark.KNOCKPY_AVAILABLE",
            True,
        ):
            result, meta = selector.select(np.zeros((1, 5)), np.array([0]))
        assert result["selected_count"] == 0
        assert meta["knockoff_backend"] == "skipped"

    def test_functional_wrapper_unavailable(self):
        with patch(
            "tabnetics.feature_selection.methods.knockoff_benchmark.KNOCKPY_AVAILABLE",
            False,
        ):
            X, y = _make_binary_data(n=50, p=10)
            with pytest.warns(RuntimeWarning):
                result, meta = knockoff_benchmark_selection(
                    X, y, fdr=0.10, random_state=42,
                )
        assert result["selected_count"] == 0

    def test_config_default(self):
        selector = KnockoffBenchmarkSelector()
        assert selector.fdr == 0.10
        assert selector.knockoff_type == "gaussian"
        assert selector.statistic == "lasso"
        assert selector.random_state is None

    def test_result_structure(self):
        X, y = _make_binary_data()
        with patch(
            "tabnetics.feature_selection.methods.knockoff_benchmark.KNOCKPY_AVAILABLE",
            False,
        ):
            selector = KnockoffBenchmarkSelector(fdr=0.20, random_state=0)
            with pytest.warns(RuntimeWarning):
                result, meta = selector.select(X, y)
        assert "selected_indices" in result
        assert "selected_count" in result
        assert "fdr" in meta

    def test_error_handling(self):
        """Simulate knockpy raising an exception."""
        X, y = _make_binary_data(n=50, p=10)
        selector = KnockoffBenchmarkSelector(fdr=0.10, random_state=42)
        with patch(
            "tabnetics.feature_selection.methods.knockoff_benchmark.KNOCKPY_AVAILABLE",
            True,
        ):
            with patch.object(
                selector, "_run_knockpy", side_effect=RuntimeError("mock error"),
            ):
                with pytest.warns(RuntimeWarning, match="knockpy failed"):
                    result, meta = selector.select(X, y)
        assert result["selected_count"] == 0
        assert meta["knockoff_backend"] == "error"
        assert "mock error" in meta["error"]


# ---------------------------------------------------------------------------
# Real knockpy tests (skip if not installed)
# ---------------------------------------------------------------------------


@pytest.mark.skipif(not KNOCKPY_AVAILABLE, reason="knockpy not installed")
class TestKnockoffReal:
    """Verify real knockpy integration."""

    def test_basic_selection(self):
        X, y = _make_binary_data(n=200, p=20, n_informative=5)
        selector = KnockoffBenchmarkSelector(fdr=0.20, random_state=42)
        result, meta = selector.select(X, y)
        assert "selected_indices" in result
        assert isinstance(result["selected_indices"], np.ndarray)
        assert meta["knockoff_backend"] == "knockpy"

    def test_high_fdr_selects_more(self):
        X, y = _make_binary_data(n=200, p=20, n_informative=5)
        sel_low = KnockoffBenchmarkSelector(fdr=0.05, random_state=42)
        sel_high = KnockoffBenchmarkSelector(fdr=0.50, random_state=42)
        r_low, _ = sel_low.select(X, y)
        r_high, _ = sel_high.select(X, y)
        # Higher FDR should generally select at least as many features.
        assert r_high["selected_count"] >= r_low["selected_count"]


# ---------------------------------------------------------------------------
# Config toggle test
# ---------------------------------------------------------------------------


class TestBenchmarkKnockoffConfig:
    """Verify benchmark_knockoff_enabled config field exists."""

    def test_config_field_default_false(self):
        from tabnetics.pipeline.pipeline import DFFSConfig
        cfg = DFFSConfig()
        assert hasattr(cfg, "benchmark_knockoff_enabled")
        assert cfg.benchmark_knockoff_enabled is False
